import os
from typing import List, Tuple
from copy import deepcopy
import sqlite3
import json
import time
import hashlib

import litellm
from filelock import FileLock

from .base import BaseLLM, LLMConfig
from ..utils.llm_utils import TextChatMessage
from ..utils.logging_utils import get_logger


logger = get_logger(__name__)


class LLM_Cache:
    def __init__(self, cache_dir: str, cache_filename):
        os.makedirs(cache_dir, exist_ok=True)
        self.cache_filepath =  os.path.join(cache_dir, f"{cache_filename}.sqlite")
        self.lock_file = self.cache_filepath + ".lock"

        self.__db_operation("""
            CREATE TABLE IF NOT EXISTS cache (
                key TEXT PRIMARY KEY,
                message TEXT,
                metadata TEXT
            )
        """, commit=True)
    
    def __db_operation(self, sql, parameters=(), commit=False, fetchone=False):
        with FileLock(self.lock_file):
            conn = sqlite3.connect(self.cache_filepath)
            c = conn.cursor()
            c.execute(sql, parameters)
            if commit:
                conn.commit()
            if fetchone:
                row = c.fetchone()
            conn.close()
            if fetchone:
                return row

    def __params_to_key(self, params):
        key_str = f"Model: {params['model']}, Temperature: {params['temperature']}, Messages: {params['messages']}"
        return hashlib.sha256(key_str.encode("utf-8")).hexdigest()

    def read(self, params):
        key = self.__params_to_key(params)
        row = self.__db_operation("SELECT message, metadata FROM cache WHERE key = ?", (key,), fetchone=True)
        if row is None:
            return None
        message, metadata_str = row
        metadata = json.loads(metadata_str)
        return message, metadata

    def write(self, params, message, metadata):
        key = self.__params_to_key(params)
        metadata_str = json.dumps(metadata)
        self.__db_operation("INSERT OR REPLACE INTO cache (key, message, metadata) VALUES (?, ?, ?)", (key, message, metadata_str), commit=True)


class BedrockLLM(BaseLLM):
    """
    To select this implementation you can initialise HippoRAG with:
        llm_model_name="anthropic.claude-3-5-haiku-20241022-v1:0" or any other Bedrock Model-ID
    """
    def __init__(self, global_config = None):
        self.global_config = global_config
        super().__init__(global_config)
        self._init_llm_config()

        self.cache = LLM_Cache(
            os.path.join(global_config.save_dir, "llm_cache"),
            self.llm_name.replace('/', '_'))        
        
        self.retry = 5
        
        logger.info(f"[BedrockLLM] Model-ID: {self.global_config.llm_name}, Cache: {self.cache.cache_filepath}")

    def _init_llm_config(self) -> None:
        config_dict = self.global_config.__dict__
        config_dict['llm_name'] = self.global_config.llm_name
        config_dict['generate_params'] = {
                "model": self.global_config.llm_name,
                "n": 1,
                "temperature": config_dict.get("temperature", 0.0),
            }

        self.llm_config = LLMConfig.from_dict(config_dict=config_dict)
        logger.info(f"[BedrockLLM] Config: {self.llm_config}")

    def __llm_call(self, params):
        num, wait_s = 0, 0.5
        while True:
            try:
                return litellm.completion(**params)
            except Exception as e:
                num += 1
                if num > self.retry:
                    raise e
                
                logger.warning(f"Bedrock LLM Exception: {e}\nRetry #{num} after {wait_s} seconds")
                time.sleep(wait_s)
                wait_s *= 2
    
    def infer(self, messages: List[TextChatMessage], **kwargs) -> Tuple[List[TextChatMessage], dict]:
        params = deepcopy(self.llm_config.generate_params)
        if kwargs:
            params.update(kwargs)
        params["messages"] = messages
        
        cache_lookup = self.cache.read(params)
        if cache_lookup is not None:
            cached = True
            message, metadata = cache_lookup
        else:
            cached = False
            response = self.__llm_call(params)
            message = response.choices[0].message.content
            metadata = {
                "prompt_tokens": response.usage.prompt_tokens, 
                "completion_tokens": response.usage.completion_tokens,
                "finish_reason": response.choices[0].finish_reason,
            }
            self.cache.write(params, message, metadata)

        return message, metadata, cached



one_shot_ircot_demo_docs = (
    """Wikipedia Title: Milk and Honey (album)\nMilk and Honey is an album by John Lennon and Yoko Ono released in 1984. Following the compilation "The John Lennon Collection", it is Lennon's eighth and final studio album, and the first posthumous release of new Lennon music, having been recorded in the last months of his life during and following the sessions for their 1980 album "Double Fantasy". It was assembled by Yoko Ono in association with the Geffen label.\n\n"""
    """Wikipedia Title: John Lennon Museum\nJohn Lennon Museum (ジョン・レノン・ミュージアム , Jon Renon Myūjiamu ) was a museum located inside the Saitama Super Arena in Chūō-ku, Saitama, Saitama Prefecture, Japan. It was established to preserve knowledge of John Lennon's life and musical career. It displayed Lennon's widow Yoko Ono's collection of his memorabilia as well as other displays. The museum opened on October 9, 2000, the 60th anniversary of Lennon’s birth, and closed on September 30, 2010, when its exhibit contract with Yoko Ono expired. A tour of the museum began with a welcoming message and short film narrated by Yoko Ono (in Japanese with English headphones available), and ended at an avant-garde styled "reflection room" full of chairs facing a slide show of moving words and images. After this room there was a gift shop with John Lennon memorabilia available.\n\n"""
    """Wikipedia Title: Walls and Bridges\nWalls and Bridges is the fifth studio album by English musician John Lennon. It was issued by Apple Records on 26 September 1974 in the United States and on 4 October in the United Kingdom. Written, recorded and released during his 18-month separation from Yoko Ono, the album captured Lennon in the midst of his "Lost Weekend". "Walls and Bridges" was an American "Billboard" number-one album and featured two hit singles, "Whatever Gets You thru the Night" and "#9 Dream". The first of these was Lennon's first number-one hit in the United States as a solo artist, and his only chart-topping single in either the US or Britain during his lifetime.\n\n"""
    """Wikipedia Title: Nobody Loves You (When You're Down and Out)\n"Nobody Loves You (When You're Down and Out)" is a song written by John Lennon released on his 1974 album "Walls and Bridges". The song is included on the 1986 compilation "Menlove Ave.", the 1990 boxset "Lennon", the 1998 boxset "John Lennon Anthology", the 2005 two-disc compilation "", and the 2010 boxset "Gimme Some Truth".\n\n"""
    """Wikipedia Title: Give Peace a Chance\n"Give Peace a Chance" is an anti-war song written by John Lennon (credited to Lennon–McCartney), and performed with Yoko Ono in Montreal, Quebec, Canada. Released as a single in 1969 by the Plastic Ono Band on Apple Records (catalogue Apple 13 in the United Kingdom, Apple 1809 in the United States), it is the first solo single issued by Lennon, released when he was still a member of the Beatles, and became an anthem of the American anti-war movement during the 1970s. It peaked at number 14 on the "Billboard" Hot 100 and number 2 on the British singles chart.\n"""
)


one_shot_ircot_demo = (
    f'{one_shot_ircot_demo_docs}'
    '\n\nQuestion: '
    f"Nobody Loves You was written by John Lennon and released on what album that was issued by Apple Records, and was written, recorded, and released during his 18 month separation from Yoko Ono?"
    '\nThought: '
    f"The album issued by Apple Records, and written, recorded, and released during John Lennon's 18 month separation from Yoko Ono is Walls and Bridges. Nobody Loves You was written by John Lennon on Walls and Bridges album. So the answer is: Walls and Bridges."
    '\n\n'
)

ircot_system = (
    'You serve as an intelligent assistant, adept at facilitating users through complex, multi-hop reasoning across multiple documents. This task is illustrated through demonstrations, each consisting of a document set paired with a relevant question and its multi-hop reasoning thoughts. Your task is to generate one thought for current step, DON\'T generate the whole thoughts at once! If you reach what you believe to be the final step, start with "So the answer is:".'
    '\n\n'
    f'{one_shot_ircot_demo}'
)


prompt_template = [
    {"role": "system", "content": ircot_system},
    {"role": "user", "content": "${prompt_user}"}
]



"""
Numba 优化工具模块

该模块使用 Numba JIT 编译器优化计算密集型函数，
特别适用于大规模向量运算、矩阵计算和数值处理场景。

优化策略：
1. 余弦相似度计算 - Numba 并行优化效果显著（10-20x 加速）
2. Min-Max 归一化 - 大规模数据有 2-5x 加速
3. 点积计算 - NumPy 已高度优化，仅在大规模场景使用 Numba

使用方法：
    from foxhipporag.utils.numba_utils import (
        numba_cosine_similarity,
        numba_min_max_normalize,
        numba_top_k_indices,
    )

注意：
- 首次调用会有 JIT 编译开销，建议在程序启动时预热
- 对于小规模数据（<1000），NumPy 可能更快
- GPU 环境优先使用 PyTorch GPU 加速
"""

import numpy as np
import logging
from typing import Tuple, Optional, List

logger = logging.getLogger(__name__)

# 尝试导入 Numba
try:
    from numba import jit, prange, float32, int64
    NUMBA_AVAILABLE = True
    logger.info("Numba JIT 编译器已启用")
except ImportError:
    NUMBA_AVAILABLE = False
    logger.warning("Numba 未安装，使用纯 Python 实现。建议: pip install numba")
    
    def jit(*args, **kwargs):
        def decorator(func):
            return func
        if len(args) == 1 and callable(args[0]):
            return args[0]
        return decorator
    
    def prange(*args, **kwargs):
        return range(*args)
    
    float32 = np.float32
    int64 = np.int64


# 最小数据规模阈值（小于此值使用 NumPy 更快）
# 不同操作有不同的最优阈值
MIN_SIZE_FOR_COSINE = 1000      # 余弦相似度：Numba 效果好
MIN_SIZE_FOR_NORMALIZE = 50000  # 归一化：NumPy 在小规模更快
MIN_SIZE_FOR_TOPK = 5000        # Top-K：中等阈值


# ============================================================
# 余弦相似度计算优化（效果最显著）
# ============================================================

@jit(nopython=True, cache=True, fastmath=True)
def _cosine_similarity_single(query: np.ndarray, matrix: np.ndarray) -> np.ndarray:
    """
    计算单个查询向量与矩阵中所有向量的余弦相似度
    
    Args:
        query: 查询向量 (dim,)
        matrix: 向量矩阵 (n, dim)
        
    Returns:
        余弦相似度 (n,)
    """
    n = matrix.shape[0]
    result = np.empty(n, dtype=np.float32)
    
    # 计算查询向量的范数
    query_norm = 0.0
    for j in range(query.shape[0]):
        query_norm += query[j] * query[j]
    query_norm = np.sqrt(query_norm)
    
    if query_norm < 1e-10:
        return np.zeros(n, dtype=np.float32)
    
    for i in range(n):
        dot = 0.0
        mat_norm = 0.0
        for j in range(query.shape[0]):
            dot += query[j] * matrix[i, j]
            mat_norm += matrix[i, j] * matrix[i, j]
        mat_norm = np.sqrt(mat_norm)
        
        if mat_norm < 1e-10:
            result[i] = 0.0
        else:
            result[i] = dot / (query_norm * mat_norm)
    
    return result


@jit(nopython=True, cache=True, fastmath=True, parallel=True)
def _cosine_similarity_batch(queries: np.ndarray, matrix: np.ndarray) -> np.ndarray:
    """
    批量计算查询向量与矩阵中所有向量的余弦相似度
    
    Args:
        queries: 查询向量矩阵 (m, dim)
        matrix: 向量矩阵 (n, dim)
        
    Returns:
        余弦相似度矩阵 (m, n)
    """
    m = queries.shape[0]
    n = matrix.shape[0]
    result = np.empty((m, n), dtype=np.float32)
    
    # 预计算矩阵向量的范数
    matrix_norms = np.empty(n, dtype=np.float32)
    for i in range(n):
        norm = 0.0
        for j in range(matrix.shape[1]):
            norm += matrix[i, j] * matrix[i, j]
        matrix_norms[i] = np.sqrt(norm)
    
    for i in prange(m):
        query_norm = 0.0
        for j in range(queries.shape[1]):
            query_norm += queries[i, j] * queries[i, j]
        query_norm = np.sqrt(query_norm)
        
        if query_norm < 1e-10:
            for j in range(n):
                result[i, j] = 0.0
            continue
        
        for j in range(n):
            dot = 0.0
            for k in range(queries.shape[1]):
                dot += queries[i, k] * matrix[j, k]
            
            if matrix_norms[j] < 1e-10:
                result[i, j] = 0.0
            else:
                result[i, j] = dot / (query_norm * matrix_norms[j])
    
    return result


def numba_cosine_similarity(
    query: np.ndarray, 
    matrix: np.ndarray,
    min_size: int = None
) -> np.ndarray:
    """
    余弦相似度计算（Numba 优化效果显著）
    
    对于大规模数据（>1000），Numba 提供显著加速（10-20x）
    
    Args:
        query: 查询向量或矩阵
        matrix: 向量矩阵
        min_size: 使用 Numba 的最小数据规模（默认 1000）
        
    Returns:
        余弦相似度结果
    """
    if min_size is None:
        min_size = MIN_SIZE_FOR_COSINE
    query = np.ascontiguousarray(query, dtype=np.float32)
    matrix = np.ascontiguousarray(matrix, dtype=np.float32)
    
    n = matrix.shape[0]
    
    # 小规模数据使用 NumPy
    if n < min_size or not NUMBA_AVAILABLE:
        return _numpy_cosine_similarity(query, matrix)
    
    if query.ndim == 1:
        return _cosine_similarity_single(query, matrix)
    else:
        return _cosine_similarity_batch(query, matrix)


def _numpy_cosine_similarity(query: np.ndarray, matrix: np.ndarray) -> np.ndarray:
    """NumPy 实现的余弦相似度（小规模数据更快）"""
    # L2 归一化
    if query.ndim == 1:
        query_norm = query / np.maximum(np.linalg.norm(query), 1e-10)
    else:
        query_norm = query / np.maximum(np.linalg.norm(query, axis=1, keepdims=True), 1e-10)
    
    matrix_norms = np.maximum(np.linalg.norm(matrix, axis=1, keepdims=True), 1e-10)
    matrix_normalized = matrix / matrix_norms
    
    return np.dot(query_norm if query.ndim == 1 else query_norm, matrix_normalized.T)


# ============================================================
# Min-Max 归一化
# 注意：NumPy 的 min/max 是高度优化的 C 实现
# Numba 仅在超大规模数据（>50000）时有优势
# ============================================================

def numba_min_max_normalize(x: np.ndarray, min_size: int = MIN_SIZE_FOR_NORMALIZE) -> np.ndarray:
    """
    Min-Max 归一化
    
    对于大多数情况，NumPy 实现已经足够快。
    Numba 版本仅在超大规模数据时有轻微优势。
    
    Args:
        x: 输入数组
        min_size: 使用 Numba 的最小数据规模（默认 50000）
        
    Returns:
        归一化后的数组
    """
    # 对于大多数实际场景，NumPy 更快
    if x.size < min_size or not NUMBA_AVAILABLE:
        return _numpy_min_max_normalize(x)
    
    # 超大规模数据使用 Numba
    x = np.ascontiguousarray(x, dtype=np.float32)
    return _min_max_normalize_1d(x.ravel()).reshape(x.shape)


def _numpy_min_max_normalize(x: np.ndarray) -> np.ndarray:
    """NumPy 实现的 Min-Max 归一化（简洁高效）"""
    min_val = np.min(x)
    max_val = np.max(x)
    range_val = max_val - min_val
    
    if range_val == 0:
        return np.ones_like(x)
    
    return (x - min_val) / range_val


@jit(nopython=True, cache=True, fastmath=True, parallel=True)
def _min_max_normalize_1d(x: np.ndarray) -> np.ndarray:
    """Min-Max 归一化 - Numba 版本（仅用于超大规模数据）"""
    n = x.shape[0]
    
    min_val = x[0]
    max_val = x[0]
    for i in range(1, n):
        if x[i] < min_val:
            min_val = x[i]
        if x[i] > max_val:
            max_val = x[i]
    
    range_val = max_val - min_val
    
    if range_val < 1e-10:
        return np.ones(n, dtype=np.float32)
    
    result = np.empty(n, dtype=np.float32)
    for i in prange(n):
        result[i] = (x[i] - min_val) / range_val
    
    return result


# ============================================================
# Top-K 选择优化
# ============================================================

@jit(nopython=True, cache=True)
def _top_k_indices_1d(scores: np.ndarray, k: int) -> Tuple[np.ndarray, np.ndarray]:
    """Top-K 选择（1D 数组）"""
    n = scores.shape[0]
    k = min(k, n)
    
    # 使用排序（Numba 不支持 argpartition）
    indices = np.argsort(scores)[::-1][:k]
    top_scores = scores[indices]
    
    return indices.astype(np.int64), top_scores


@jit(nopython=True, cache=True, parallel=True)
def _top_k_indices_2d(scores: np.ndarray, k: int) -> Tuple[np.ndarray, np.ndarray]:
    """Top-K 选择（2D 数组）"""
    m = scores.shape[0]
    n = scores.shape[1]
    k = min(k, n)
    
    indices = np.empty((m, k), dtype=np.int64)
    top_scores = np.empty((m, k), dtype=np.float32)
    
    for i in prange(m):
        row_indices = np.argsort(scores[i])[::-1][:k]
        for j in range(k):
            indices[i, j] = row_indices[j]
            top_scores[i, j] = scores[i, row_indices[j]]
    
    return indices, top_scores


def numba_top_k_indices(
    scores: np.ndarray, 
    k: int,
    min_size: int = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Top-K 选择
    
    Args:
        scores: 得分数组或矩阵
        k: 要选择的数量
        min_size: 使用 Numba 的最小数据规模（默认 5000）
        
    Returns:
        (indices, scores): Top-K 索引和得分
    """
    if min_size is None:
        min_size = MIN_SIZE_FOR_TOPK
    
    scores = np.ascontiguousarray(scores, dtype=np.float32)
    
    # 小规模数据使用 NumPy
    if scores.size < min_size or not NUMBA_AVAILABLE:
        return _numpy_top_k_indices(scores, k)
    
    if scores.ndim == 1:
        return _top_k_indices_1d(scores, k)
    else:
        return _top_k_indices_2d(scores, k)


def _numpy_top_k_indices(scores: np.ndarray, k: int) -> Tuple[np.ndarray, np.ndarray]:
    """NumPy 实现的 Top-K 选择"""
    if scores.ndim == 1:
        indices = np.argsort(scores)[::-1][:k]
        return indices.astype(np.int64), scores[indices]
    else:
        indices = np.argsort(scores, axis=1)[:, ::-1][:, :k]
        top_scores = np.take_along_axis(scores, indices, axis=1)
        return indices.astype(np.int64), top_scores


# ============================================================
# L2 归一化
# ============================================================

@jit(nopython=True, cache=True, fastmath=True, parallel=True)
def _l2_normalize_2d(vectors: np.ndarray) -> np.ndarray:
    """L2 归一化（2D 数组）"""
    n = vectors.shape[0]
    dim = vectors.shape[1]
    result = np.empty((n, dim), dtype=np.float32)
    
    for i in prange(n):
        norm = 0.0
        for j in range(dim):
            norm += vectors[i, j] * vectors[i, j]
        norm = np.sqrt(norm)
        
        if norm < 1e-10:
            for j in range(dim):
                result[i, j] = 0.0
        else:
            for j in range(dim):
                result[i, j] = vectors[i, j] / norm
    
    return result


def numba_l2_normalize(vectors: np.ndarray, min_size: int = None) -> np.ndarray:
    """
    L2 归一化
    
    Args:
        vectors: 输入向量或矩阵
        min_size: 使用 Numba 的最小数据规模（默认 1000）
        
    Returns:
        归一化后的向量或矩阵
    """
    if min_size is None:
        min_size = MIN_SIZE_FOR_COSINE
    vectors = np.ascontiguousarray(vectors, dtype=np.float32)
    
    # 小规模数据使用 NumPy
    if vectors.size < min_size or not NUMBA_AVAILABLE:
        norms = np.linalg.norm(vectors, axis=-1, keepdims=True)
        return vectors / np.maximum(norms, 1e-10)
    
    if vectors.ndim == 1:
        vectors = vectors.reshape(1, -1)
        return _l2_normalize_2d(vectors)[0]
    else:
        return _l2_normalize_2d(vectors)


# ============================================================
# KNN 检索（使用优化的余弦相似度）
# ============================================================

def numba_knn_search(
    query: np.ndarray,
    index_vectors: np.ndarray,
    k: int = 10
) -> Tuple[np.ndarray, np.ndarray]:
    """
    使用 Numba 优化的 KNN 检索
    
    Args:
        query: 查询向量或矩阵
        index_vectors: 索引向量矩阵
        k: 返回的最近邻数量
        
    Returns:
        (indices, scores): 最近邻索引和相似度得分
    """
    query = np.ascontiguousarray(query, dtype=np.float32)
    index_vectors = np.ascontiguousarray(index_vectors, dtype=np.float32)
    
    # 计算相似度
    scores = numba_cosine_similarity(query, index_vectors)
    
    # Top-K 选择
    return numba_top_k_indices(scores, k)


# ============================================================
# 工具函数
# ============================================================

def is_numba_available() -> bool:
    """检查 Numba 是否可用"""
    return NUMBA_AVAILABLE


def get_numba_info() -> dict:
    """获取 Numba 相关信息"""
    info = {
        'numba_available': NUMBA_AVAILABLE,
        'thresholds': {
            'cosine_similarity': MIN_SIZE_FOR_COSINE,
            'min_max_normalize': MIN_SIZE_FOR_NORMALIZE,
            'top_k': MIN_SIZE_FOR_TOPK,
        }
    }
    
    if NUMBA_AVAILABLE:
        try:
            import numba
            info['numba_version'] = numba.__version__
            info['cuda_available'] = numba.cuda.is_available()
        except Exception as e:
            info['error'] = str(e)
    
    return info


def warmup():
    """
    预热 Numba JIT 编译器
    
    建议在程序启动时调用，避免首次使用时的编译延迟
    """
    if not NUMBA_AVAILABLE:
        return
    
    logger.info("预热 Numba JIT 编译器...")
    
    # 使用超过阈值的数据预热
    vec = np.random.randn(2000).astype(np.float32)
    matrix = np.random.randn(2000, 128).astype(np.float32)
    
    # 预热各函数（强制使用 Numba）
    _ = numba_cosine_similarity(vec[:128], matrix, min_size=0)
    _ = numba_min_max_normalize(vec, min_size=0)
    _ = numba_top_k_indices(vec, 10, min_size=0)
    _ = numba_l2_normalize(matrix)
    
    logger.info("Numba JIT 预热完成")


# ============================================================
# 性能测试
# ============================================================

def benchmark(size: int = 10000, dim: int = 512) -> dict:
    """
    性能基准测试
    
    Args:
        size: 向量数量
        dim: 向量维度
        
    Returns:
        性能比较结果
    """
    import time
    
    np.random.seed(42)
    matrix = np.random.randn(size, dim).astype(np.float32)
    query = np.random.randn(dim).astype(np.float32)
    
    results = {}
    
    # 余弦相似度测试
    start = time.time()
    for _ in range(5):
        _ = _numpy_cosine_similarity(query, matrix)
    numpy_time = time.time() - start
    
    start = time.time()
    for _ in range(5):
        _ = numba_cosine_similarity(query, matrix)
    numba_time = time.time() - start
    
    results['cosine_similarity'] = {
        'numpy_time': numpy_time,
        'numba_time': numba_time,
        'speedup': numpy_time / numba_time if numba_time > 0 else 0
    }
    
    # 归一化测试
    scores = np.random.randn(size).astype(np.float32)
    
    start = time.time()
    for _ in range(50):
        _ = _numpy_min_max_normalize(scores)
    numpy_time = time.time() - start
    
    start = time.time()
    for _ in range(50):
        _ = numba_min_max_normalize(scores)
    numba_time = time.time() - start
    
    results['min_max_normalize'] = {
        'numpy_time': numpy_time,
        'numba_time': numba_time,
        'speedup': numpy_time / numba_time if numba_time > 0 else 0
    }
    
    return results


# ============================================================
# 加权平均计算优化
# ============================================================

@jit(nopython=True, cache=True, parallel=True)
def numba_weighted_average(values: np.ndarray, weights: np.ndarray) -> float:
    """
    计算加权平均
    
    Args:
        values: 值数组
        weights: 权重数组
        
    Returns:
        加权平均值
    """
    total_weight = 0.0
    weighted_sum = 0.0
    
    for i in prange(len(values)):
        total_weight += weights[i]
        weighted_sum += values[i] * weights[i]
    
    if total_weight < 1e-10:
        return 0.0
    
    return weighted_sum / total_weight


@jit(nopython=True, cache=True, parallel=True)
def numba_batch_weighted_average(values_matrix: np.ndarray, weights_matrix: np.ndarray) -> np.ndarray:
    """
    批量计算加权平均
    
    Args:
        values_matrix: 值矩阵 (n, m)
        weights_matrix: 权重矩阵 (n, m)
        
    Returns:
        加权平均值数组 (n,)
    """
    n = values_matrix.shape[0]
    result = np.empty(n, dtype=np.float32)
    
    for i in prange(n):
        total_weight = 0.0
        weighted_sum = 0.0
        
        for j in range(values_matrix.shape[1]):
            total_weight += weights_matrix[i, j]
            weighted_sum += values_matrix[i, j] * weights_matrix[i, j]
        
        if total_weight < 1e-10:
            result[i] = 0.0
        else:
            result[i] = weighted_sum / total_weight
    
    return result


# ============================================================
# 稀疏向量操作优化
# ============================================================

@jit(nopython=True, cache=True)
def numba_sparse_dot_product(indices1: np.ndarray, values1: np.ndarray,
                              indices2: np.ndarray, values2: np.ndarray) -> float:
    """
    计算两个稀疏向量的点积
    
    Args:
        indices1: 第一个向量的非零索引
        values1: 第一个向量的非零值
        indices2: 第二个向量的非零索引
        values2: 第二个向量的非零值
        
    Returns:
        点积结果
    """
    result = 0.0
    i, j = 0, 0
    
    while i < len(indices1) and j < len(indices2):
        if indices1[i] == indices2[j]:
            result += values1[i] * values2[j]
            i += 1
            j += 1
        elif indices1[i] < indices2[j]:
            i += 1
        else:
            j += 1
    
    return result


# ============================================================
# 分数融合优化
# ============================================================

@jit(nopython=True, cache=True, parallel=True)
def numba_fuse_scores(scores1: np.ndarray, scores2: np.ndarray, 
                      weight1: float, weight2: float) -> np.ndarray:
    """
    融合两组分数
    
    Args:
        scores1: 第一组分数
        scores2: 第二组分数
        weight1: 第一组权重
        weight2: 第二组权重
        
    Returns:
        融合后的分数
    """
    n = len(scores1)
    result = np.empty(n, dtype=np.float32)
    
    for i in prange(n):
        result[i] = weight1 * scores1[i] + weight2 * scores2[i]
    
    return result


@jit(nopython=True, cache=True, parallel=True)
def numba_multiplicative_fuse(scores1: np.ndarray, scores2: np.ndarray,
                               alpha: float = 0.5) -> np.ndarray:
    """
    乘法融合两组分数
    
    结合加权和乘法融合，确保两者都高时得分才高
    
    Args:
        scores1: 第一组分数
        scores2: 第二组分数
        alpha: 加权融合的权重
        
    Returns:
        融合后的分数
    """
    n = len(scores1)
    result = np.empty(n, dtype=np.float32)
    
    for i in prange(n):
        multiplicative = scores1[i] * scores2[i]
        weighted = alpha * scores1[i] + (1 - alpha) * scores2[i]
        result[i] = 0.3 * multiplicative + 0.7 * weighted
    
    return result


# ============================================================
# 批量哈希计算优化（用于实体ID计算）
# ============================================================

def batch_compute_mdhash_id(contents: List[str], prefix: str = "") -> List[str]:
    """
    批量计算MD5哈希ID
    
    使用多线程加速批量哈希计算
    
    Args:
        contents: 内容字符串列表
        prefix: 前缀
        
    Returns:
        哈希ID列表
    """
    import hashlib
    from hashlib import md5
    
    result = []
    for content in contents:
        hash_id = prefix + md5(content.encode()).hexdigest()
        result.append(hash_id)
    
    return result


# ============================================================
# 索引映射优化
# ============================================================

@jit(nopython=True, cache=True)
def numba_build_index_mapping(keys: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    构建索引映射
    
    Args:
        keys: 键数组
        
    Returns:
        (unique_keys, indices): 唯一键和对应的索引
    """
    n = len(keys)
    unique_keys = []
    indices = np.zeros(n, dtype=np.int64)
    
    key_to_idx = {}
    current_idx = 0
    
    for i in range(n):
        key = keys[i]
        if key not in key_to_idx:
            key_to_idx[key] = current_idx
            unique_keys.append(key)
            current_idx += 1
        indices[i] = key_to_idx[key]
    
    return np.array(unique_keys), indices


# ============================================================
# 批量 Top-K 选择优化（支持动态 K）
# ============================================================

@jit(nopython=True, cache=True)
def numba_batch_top_k(scores_matrix: np.ndarray, k: int) -> Tuple[np.ndarray, np.ndarray]:
    """
    批量 Top-K 选择
    
    Args:
        scores_matrix: 得分矩阵 (n, m)
        k: 每行选择的 top-k 数量
        
    Returns:
        (indices, scores): Top-K 索引和得分
    """
    n = scores_matrix.shape[0]
    m = scores_matrix.shape[1]
    k = min(k, m)
    
    all_indices = np.empty((n, k), dtype=np.int64)
    all_scores = np.empty((n, k), dtype=np.float32)
    
    for i in range(n):
        row = scores_matrix[i]
        sorted_indices = np.argsort(row)[::-1][:k]
        
        for j in range(k):
            all_indices[i, j] = sorted_indices[j]
            all_scores[i, j] = row[sorted_indices[j]]
    
    return all_indices, all_scores


if __name__ == '__main__':
    print("Numba 优化工具模块")
    print("=" * 50)
    
    info = get_numba_info()
    print(f"Numba 可用: {info['numba_available']}")
    
    if info['numba_available']:
        print(f"Numba 版本: {info.get('numba_version', 'unknown')}")
        print(f"CUDA 可用: {info.get('cuda_available', False)}")
        
        # 预热
        warmup()
        
        # 运行基准测试
        print("\n运行性能基准测试...")
        results = benchmark(size=10000, dim=512)
        
        print("\n性能比较结果:")
        for test_name, data in results.items():
            print(f"\n{test_name}:")
            print(f"  NumPy 时间: {data['numpy_time']:.4f}s")
            print(f"  Numba 时间: {data['numba_time']:.4f}s")
            print(f"  加速比: {data['speedup']:.2f}x")



import json
from typing import Dict, Tuple

from ..information_extraction import OpenIE
from .openie_openai import ChunkInfo
from ..utils.misc_utils import NerRawOutput, TripleRawOutput
from ..utils.logging_utils import get_logger
from ..prompts import PromptTemplateManager
from ..llm.transformers_offline import TransformersOffline

logger = get_logger(__name__)


class TransformersOfflineOpenIE(OpenIE):
    def __init__(self, global_config):

        self.prompt_template_manager = PromptTemplateManager(role_mapping={"system": "system", "user": "user", "assistant": "assistant"})
        self.llm_model = TransformersOffline(global_config)

    def batch_openie(self, chunks: Dict[str, ChunkInfo]) -> Tuple[Dict[str, NerRawOutput], Dict[str, TripleRawOutput]]:
        """
        Conduct batch OpenIE synchronously using vLLM offline batch mode, including NER and triple extraction

        Args:
            chunks (Dict[str, ChunkInfo]): chunks to be incorporated into graph. Each key is a hashed chunk
            and the corresponding value is the chunk info to insert.

        Returns:
            Tuple[Dict[str, NerRawOutput], Dict[str, TripleRawOutput]]:
                - A dict with keys as the chunk ids and values as the NER result instances.
                - A dict with keys as the chunk ids and values as the triple extraction result instances.
        """

        # Extract passages from the provided chunks
        chunk_passages = {chunk_key: chunk["content"] for chunk_key, chunk in chunks.items()}

        ner_input_messages = [self.prompt_template_manager.render(name='ner', passage=p) for p in chunk_passages.values()]
        ner_output, ner_output_metadata = self.llm_model.batch_infer(ner_input_messages, json_template='ner', max_tokens=512)

        triple_extract_input_messages = [self.prompt_template_manager.render(
            name='triple_extraction',
            passage=passage,
            named_entity_json=named_entities
        ) for passage, named_entities in zip(chunk_passages.values(), ner_output)]
        triple_output, triple_output_metadata = self.llm_model.batch_infer(triple_extract_input_messages, json_template='triples', max_tokens=2048)

        ner_raw_outputs = []
        for idx, ner_output_instance in enumerate(ner_output):
            chunk_id = list(chunks.keys())[idx]
            response = ner_output_instance
            try:
                unique_entities = json.loads(response)["named_entities"]
            except Exception as e:
                unique_entities = []
                logger.warning(f"Could not parse response from OpenIE: {e}")
            if len(unique_entities) == 0:
                logger.warning("No entities extracted for chunk_id: {}".format(chunk_id))
            ner_raw_output = NerRawOutput(chunk_id, response, unique_entities, {})
            ner_raw_outputs.append(ner_raw_output)
        ner_results_dict = {chunk_key: ner_raw_output for chunk_key, ner_raw_output in zip(chunks.keys(), ner_raw_outputs)}

        triple_raw_outputs = []
        for idx, triple_output_instance in enumerate(triple_output):
            chunk_id = list(chunks.keys())[idx]
            response = triple_output_instance
            try:
                triples = json.loads(response)["triples"]
            except Exception as e:
                triples = []
                logger.warning(f"Could not parse response from OpenIE: {e}")
            if len(triples) == 0:
                logger.warning("No triples extracted for chunk_id: {}".format(chunk_id))
            triple_raw_output = TripleRawOutput(chunk_id, response, triples, {})
            triple_raw_outputs.append(triple_raw_output)
        triple_results_dict = {chunk_key: triple_raw_output for chunk_key, triple_raw_output in zip(chunks.keys(), triple_raw_outputs)}

        return ner_results_dict, triple_results_dict



"""
Rust 加速工具模块

该模块提供高性能的 Rust 实现，用于加速 foxHippoRAG 的核心计算。

模块结构：
- similarity: 余弦相似度计算
- topk: Top-K 选择算法
- normalize: 向量归一化
- knn: KNN 检索
- text: 文本特征计算
- density: 段落密度评估

使用方法：
    from foxhipporag.utils.rust_utils import (
        cosine_similarity,
        top_k_indices,
        knn_search,
        batch_compute_density,
    )

性能优势：
- 余弦相似度：2-5x 加速
- Top-K 选择：3-10x 加速
- KNN 检索：2-4x 加速
- 文本特征计算：5-20x 加速
"""

import logging
import os
from typing import List, Dict, Tuple, Optional, Union

import numpy as np

logger = logging.getLogger(__name__)

# 尝试导入 Rust 扩展
RUST_AVAILABLE = False
try:
    import foxhipporag_rust
    RUST_AVAILABLE = True
    logger.info("foxhipporag_rust Rust 扩展已启用")
except ImportError:
    logger.debug("foxhipporag_rust 未安装，使用 Python 回退实现")


def is_rust_available() -> bool:
    """检查 Rust 扩展是否可用"""
    return RUST_AVAILABLE


def get_backend_info() -> Dict:
    """获取后端信息"""
    info = {
        "rust_available": RUST_AVAILABLE,
        "backend": "rust" if RUST_AVAILABLE else "python",
    }
    
    if RUST_AVAILABLE:
        try:
            info["rust_version"] = foxhipporag_rust.__version__
            info["rust_info"] = foxhipporag_rust.get_rust_info()
            info["num_threads"] = foxhipporag_rust.get_num_threads()
        except Exception as e:
            info["error"] = str(e)
    
    return info


# ============================================================
# 余弦相似度计算
# ============================================================

def cosine_similarity(
    query: np.ndarray,
    matrix: np.ndarray,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    计算单个查询向量与矩阵中所有向量的余弦相似度
    
    Args:
        query: 查询向量 (dim,)
        matrix: 向量矩阵 (n, dim)
        use_rust: 是否使用 Rust 加速（None 表示自动检测）
        
    Returns:
        余弦相似度数组 (n,)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    query = np.ascontiguousarray(query, dtype=np.float32)
    matrix = np.ascontiguousarray(matrix, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.cosine_similarity(query, matrix)
        except Exception as e:
            logger.warning(f"Rust cosine_similarity 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    query_norm = np.linalg.norm(query)
    if query_norm < 1e-10:
        return np.zeros(matrix.shape[0], dtype=np.float32)
    
    matrix_norms = np.linalg.norm(matrix, axis=1)
    matrix_norms = np.maximum(matrix_norms, 1e-10)
    
    return np.dot(matrix, query) / (matrix_norms * query_norm)


def cosine_similarity_batch(
    queries: np.ndarray,
    keys: np.ndarray,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    批量计算查询向量矩阵与键向量矩阵的余弦相似度
    
    Args:
        queries: 查询向量矩阵 (m, dim)
        keys: 键向量矩阵 (n, dim)
        use_rust: 是否使用 Rust 加速
        
    Returns:
        相似度矩阵 (m, n)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    queries = np.ascontiguousarray(queries, dtype=np.float32)
    keys = np.ascontiguousarray(keys, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.cosine_similarity_batch(queries, keys)
        except Exception as e:
            logger.warning(f"Rust cosine_similarity_batch 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    query_norms = np.linalg.norm(queries, axis=1, keepdims=True)
    query_norms = np.maximum(query_norms, 1e-10)
    queries_normalized = queries / query_norms
    
    key_norms = np.linalg.norm(keys, axis=1, keepdims=True)
    key_norms = np.maximum(key_norms, 1e-10)
    keys_normalized = keys / key_norms
    
    return np.dot(queries_normalized, keys_normalized.T)


# ============================================================
# Top-K 选择
# ============================================================

def top_k_indices(
    scores: np.ndarray,
    k: int,
    use_rust: Optional[bool] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    从一维数组中选择 Top-K 元素
    
    Args:
        scores: 得分数组 (n,)
        k: 要选择的数量
        use_rust: 是否使用 Rust 加速
        
    Returns:
        (indices, scores): Top-K 索引和得分
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    scores = np.ascontiguousarray(scores, dtype=np.float32)
    k = min(k, len(scores))
    
    if use_rust:
        try:
            return foxhipporag_rust.top_k_indices(scores, k)
        except Exception as e:
            logger.warning(f"Rust top_k_indices 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    indices = np.argsort(scores)[::-1][:k]
    return indices.astype(np.int64), scores[indices]


def top_k_indices_2d(
    scores: np.ndarray,
    k: int,
    use_rust: Optional[bool] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    从二维矩阵的每一行选择 Top-K 元素
    
    Args:
        scores: 得分矩阵 (m, n)
        k: 每行要选择的数量
        use_rust: 是否使用 Rust 加速
        
    Returns:
        (indices, scores): Top-K 索引和得分矩阵 (m, k)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    scores = np.ascontiguousarray(scores, dtype=np.float32)
    k = min(k, scores.shape[1])
    
    if use_rust:
        try:
            return foxhipporag_rust.top_k_indices_2d(scores, k)
        except Exception as e:
            logger.warning(f"Rust top_k_indices_2d 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    indices = np.argsort(scores, axis=1)[:, ::-1][:, :k]
    top_scores = np.take_along_axis(scores, indices, axis=1)
    return indices.astype(np.int64), top_scores


# ============================================================
# 归一化
# ============================================================

def l2_normalize(
    vector: np.ndarray,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    L2 归一化向量
    
    Args:
        vector: 输入向量
        use_rust: 是否使用 Rust 加速
        
    Returns:
        归一化后的向量
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    vector = np.ascontiguousarray(vector, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.l2_normalize(vector)
        except Exception as e:
            logger.warning(f"Rust l2_normalize 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    norm = np.linalg.norm(vector)
    if norm < 1e-10:
        return np.zeros_like(vector)
    return vector / norm


def min_max_normalize(
    values: np.ndarray,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    Min-Max 归一化
    
    Args:
        values: 输入数组
        use_rust: 是否使用 Rust 加速
        
    Returns:
        归一化后的数组
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    values = np.ascontiguousarray(values, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.min_max_normalize(values)
        except Exception as e:
            logger.warning(f"Rust min_max_normalize 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    min_val = np.min(values)
    max_val = np.max(values)
    range_val = max_val - min_val
    
    if range_val < 1e-10:
        return np.ones_like(values)
    
    return (values - min_val) / range_val


def batch_l2_normalize(
    matrix: np.ndarray,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    批量 L2 归一化二维矩阵（每行独立归一化）
    
    Args:
        matrix: 输入矩阵 (m, n)
        use_rust: 是否使用 Rust 加速
        
    Returns:
        归一化后的矩阵 (m, n)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    matrix = np.ascontiguousarray(matrix, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.batch_l2_normalize(matrix)
        except Exception as e:
            logger.warning(f"Rust batch_l2_normalize 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    norms = np.linalg.norm(matrix, axis=1, keepdims=True)
    norms = np.maximum(norms, 1e-10)
    return matrix / norms


# ============================================================
# KNN 检索
# ============================================================

def knn_search(
    query: np.ndarray,
    index_vectors: np.ndarray,
    k: int = 10,
    use_rust: Optional[bool] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    KNN 检索
    
    Args:
        query: 查询向量 (dim,)
        index_vectors: 索引向量矩阵 (n, dim)
        k: 返回的最近邻数量
        use_rust: 是否使用 Rust 加速
        
    Returns:
        (indices, scores): 最近邻索引和相似度得分
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    query = np.ascontiguousarray(query, dtype=np.float32)
    index_vectors = np.ascontiguousarray(index_vectors, dtype=np.float32)
    
    # 确保查询向量是 2D
    if query.ndim == 1:
        query = query.reshape(1, -1)
    
    if use_rust:
        try:
            return foxhipporag_rust.knn_search(query, index_vectors, k)
        except Exception as e:
            logger.warning(f"Rust knn_search 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    similarities = cosine_similarity(query.flatten(), index_vectors, use_rust=False)
    return top_k_indices(similarities, k, use_rust=False)


def knn_search_batch(
    queries: np.ndarray,
    index_vectors: np.ndarray,
    k: int = 10,
    use_rust: Optional[bool] = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    批量 KNN 检索
    
    Args:
        queries: 查询向量矩阵 (m, dim)
        index_vectors: 索引向量矩阵 (n, dim)
        k: 每个查询返回的最近邻数量
        use_rust: 是否使用 Rust 加速
        
    Returns:
        (indices, scores): Top-K 索引和得分矩阵 (m, k)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    queries = np.ascontiguousarray(queries, dtype=np.float32)
    index_vectors = np.ascontiguousarray(index_vectors, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.knn_search_batch(queries, index_vectors, k)
        except Exception as e:
            logger.warning(f"Rust knn_search_batch 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    similarity_matrix = cosine_similarity_batch(queries, index_vectors, use_rust=False)
    return top_k_indices_2d(similarity_matrix, k, use_rust=False)


def retrieve_knn(
    query_ids: List[str],
    key_ids: List[str],
    query_vecs: np.ndarray,
    key_vecs: np.ndarray,
    k: int = 2047,
    query_batch_size: int = 1000,
    key_batch_size: int = 10000,
    use_rust: Optional[bool] = None
) -> Dict:
    """
    完整的 KNN 检索接口
    
    Args:
        query_ids: 查询向量 ID 列表
        key_ids: 键向量 ID 列表
        query_vecs: 查询向量矩阵 (m, dim)
        key_vecs: 键向量矩阵 (n, dim)
        k: 返回的最近邻数量
        query_batch_size: 查询批处理大小
        key_batch_size: 键批处理大小
        use_rust: 是否使用 Rust 加速
        
    Returns:
        Dict: 查询 ID 到 (key_ids, scores) 的映射
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    query_vecs = np.ascontiguousarray(query_vecs, dtype=np.float32)
    key_vecs = np.ascontiguousarray(key_vecs, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.retrieve_knn(
                query_ids, key_ids, query_vecs, key_vecs,
                k, query_batch_size, key_batch_size
            )
        except Exception as e:
            logger.warning(f"Rust retrieve_knn 失败，回退到 Python: {e}")
    
    # Python 回退实现（使用现有的 embed_utils）
    from .embed_utils import retrieve_knn as py_retrieve_knn
    return py_retrieve_knn(
        query_ids, key_ids, query_vecs, key_vecs,
        k, query_batch_size, key_batch_size
    )


# ============================================================
# 文本特征计算
# ============================================================

def count_entities(text: str, use_rust: Optional[bool] = None) -> int:
    """
    计算文本中的实体数量
    
    Args:
        text: 输入文本
        use_rust: 是否使用 Rust 加速
        
    Returns:
        实体数量
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.count_entities(text)
        except Exception as e:
            logger.warning(f"Rust count_entities 失败，回退到 Python: {e}")
    
    # Python 回退实现
    import re
    patterns = [
        re.compile(r'\b[A-Z][a-z]+(?:\s+[A-Z][a-z]+)*\b'),
        re.compile(r'\b\d+(?:\.\d+)?(?:\s*(?:年|月|日|时|分|秒|km|m|kg|元|美元|亿|万))?\b'),
        re.compile(r'\b[A-Z]{2,}(?:\s+[A-Z]{2,})*\b'),
    ]
    
    count = 0
    for pattern in patterns:
        count += len(pattern.findall(text))
    return count


def count_facts(text: str, use_rust: Optional[bool] = None) -> int:
    """
    估算文本中的事实/关系数量
    
    Args:
        text: 输入文本
        use_rust: 是否使用 Rust 加速
        
    Returns:
        估算的事实数量
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.count_facts(text)
        except Exception as e:
            logger.warning(f"Rust count_facts 失败，回退到 Python: {e}")
    
    # Python 回退实现
    import re
    cn_patterns = ['是', '有', '位于', '属于', '包含', '包括', '成立于', '创建于',
                   '发明', '发现', '生产', '制造', '设计', '开发', '建立', '设立']
    en_patterns = [r'\bis\b', r'\bare\b', r'\bwas\b', r'\bwere\b',
                   r'\blocated\b', r'\bbelongs\b', r'\bcontains\b', r'\bincludes\b',
                   r'\bfounded\b', r'\bcreated\b', r'\binvented\b', r'\bdiscovered\b',
                   r'\bproduced\b', r'\bmanufactured\b', r'\bdesigned\b', r'\bdeveloped\b']
    
    count = 0
    for pattern in cn_patterns:
        count += len(re.findall(pattern, text))
    for pattern in en_patterns:
        count += len(re.findall(pattern, text, re.IGNORECASE))
    return count


def compute_content_richness(text: str, use_rust: Optional[bool] = None) -> float:
    """
    计算内容丰富度
    
    Args:
        text: 输入文本
        use_rust: 是否使用 Rust 加速
        
    Returns:
        内容丰富度得分 (0-1)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.compute_content_richness(text)
        except Exception as e:
            logger.warning(f"Rust compute_content_richness 失败，回退到 Python: {e}")
    
    # Python 回退实现
    import re
    stopwords = {
        '的', '了', '是', '在', '有', '和', '与', '或', '等', '这', '那',
        'the', 'a', 'an', 'is', 'are', 'was', 'were', 'be', 'been', 'being',
        'have', 'has', 'had', 'do', 'does', 'did', 'will', 'would', 'could',
        'should', 'may', 'might', 'must', 'shall', 'can', 'need', 'dare',
        'to', 'of', 'in', 'for', 'on', 'with', 'at', 'by', 'from', 'as',
        'into', 'through', 'during', 'before', 'after', 'above', 'below',
    }
    
    words = re.findall(r'\b\w+\b', text.lower())
    if not words:
        return 0.0
    
    non_stopword_count = sum(1 for w in words if w not in stopwords)
    non_stopword_ratio = non_stopword_count / len(words)
    
    unique_words = set(words)
    vocabulary_diversity = len(unique_words) / len(words)
    
    return min(1.0, 0.6 * non_stopword_ratio + 0.4 * vocabulary_diversity)


def batch_count_entities(texts: List[str], use_rust: Optional[bool] = None) -> List[int]:
    """
    批量计算实体数量
    
    Args:
        texts: 文本列表
        use_rust: 是否使用 Rust 加速
        
    Returns:
        实体数量列表
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.batch_count_entities(texts)
        except Exception as e:
            logger.warning(f"Rust batch_count_entities 失败，回退到 Python: {e}")
    
    return [count_entities(t, use_rust=False) for t in texts]


def batch_count_facts(texts: List[str], use_rust: Optional[bool] = None) -> List[int]:
    """
    批量计算事实数量
    
    Args:
        texts: 文本列表
        use_rust: 是否使用 Rust 加速
        
    Returns:
        事实数量列表
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.batch_count_facts(texts)
        except Exception as e:
            logger.warning(f"Rust batch_count_facts 失败，回退到 Python: {e}")
    
    return [count_facts(t, use_rust=False) for t in texts]


# ============================================================
# 段落密度评估
# ============================================================

def compute_density_score(
    text: str,
    entity_weight: float = 0.3,
    fact_weight: float = 0.3,
    length_weight: float = 0.2,
    richness_weight: float = 0.2,
    use_rust: Optional[bool] = None
) -> float:
    """
    计算单个段落的密度得分
    
    Args:
        text: 段落文本
        entity_weight: 实体数量权重
        fact_weight: 事实数量权重
        length_weight: 文本长度权重
        richness_weight: 内容丰富度权重
        use_rust: 是否使用 Rust 加速
        
    Returns:
        密度得分 (0-1)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.compute_density_score(
                text, entity_weight, fact_weight, length_weight, richness_weight
            )
        except Exception as e:
            logger.warning(f"Rust compute_density_score 失败，回退到 Python: {e}")
    
    # Python 回退实现
    entity_count = count_entities(text, use_rust=False)
    fact_count = count_facts(text, use_rust=False)
    text_length_score = _compute_text_length_score(text)
    content_richness = compute_content_richness(text, use_rust=False)
    
    normalized_entity_score = min(1.0, np.log1p(entity_count) / 3.0)
    normalized_fact_score = min(1.0, np.log1p(fact_count) / 2.5)
    
    density_score = (
        entity_weight * normalized_entity_score +
        fact_weight * normalized_fact_score +
        length_weight * text_length_score +
        richness_weight * content_richness
    )
    
    return density_score


def _compute_text_length_score(text: str) -> float:
    """计算文本长度得分"""
    length = len(text)
    optimal_length = 350
    max_length = 1000
    
    if length == 0:
        return 0.0
    elif length <= optimal_length:
        return length / optimal_length
    else:
        return max(0.5, 1.0 - (length - optimal_length) / (max_length - optimal_length) * 0.5)


def batch_compute_density(
    texts: List[str],
    entity_weight: float = 0.3,
    fact_weight: float = 0.3,
    length_weight: float = 0.2,
    richness_weight: float = 0.2,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    批量计算段落密度得分
    
    Args:
        texts: 段落文本列表
        entity_weight: 实体数量权重
        fact_weight: 事实数量权重
        length_weight: 文本长度权重
        richness_weight: 内容丰富度权重
        use_rust: 是否使用 Rust 加速
        
    Returns:
        密度得分数组
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.batch_compute_density(
                texts, entity_weight, fact_weight, length_weight, richness_weight
            )
        except Exception as e:
            logger.warning(f"Rust batch_compute_density 失败，回退到 Python: {e}")
    
    # Python 回退实现
    scores = [
        compute_density_score(t, entity_weight, fact_weight, length_weight, richness_weight, use_rust=False)
        for t in texts
    ]
    return np.array(scores, dtype=np.float32)


def compute_evidence_score(
    semantic_score: float,
    density_score: float,
    query_type: str = 'general',
    use_rust: Optional[bool] = None
) -> float:
    """
    计算证据质量得分
    
    Args:
        semantic_score: 语义相似度得分 (0-1)
        density_score: 信息密度得分 (0-1)
        query_type: 查询类型 ('general', 'factual', 'exploratory')
        use_rust: 是否使用 Rust 加速
        
    Returns:
        证据质量得分 (0-1)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    if use_rust:
        try:
            return foxhipporag_rust.compute_evidence_score(semantic_score, density_score, query_type)
        except Exception as e:
            logger.warning(f"Rust compute_evidence_score 失败，回退到 Python: {e}")
    
    # Python 回退实现
    if query_type == 'factual':
        semantic_weight, density_weight = 0.7, 0.3
    elif query_type == 'exploratory':
        semantic_weight, density_weight = 0.4, 0.6
    else:
        semantic_weight, density_weight = 0.5, 0.5
    
    multiplicative_score = semantic_score * density_score
    weighted_score = semantic_weight * semantic_score + density_weight * density_score
    
    return 0.3 * multiplicative_score + 0.7 * weighted_score


def batch_evidence_scores(
    semantic_scores: np.ndarray,
    density_scores: np.ndarray,
    query_type: str = 'general',
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    批量计算证据质量得分
    
    Args:
        semantic_scores: 语义相似度得分数组 (n,)
        density_scores: 信息密度得分数组 (n,)
        query_type: 查询类型
        use_rust: 是否使用 Rust 加速
        
    Returns:
        证据质量得分数组 (n,)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    semantic_scores = np.ascontiguousarray(semantic_scores, dtype=np.float32)
    density_scores = np.ascontiguousarray(density_scores, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.batch_evidence_scores(semantic_scores, density_scores, query_type)
        except Exception as e:
            logger.warning(f"Rust batch_evidence_scores 失败，回退到 NumPy: {e}")
    
    # NumPy 回退实现
    if query_type == 'factual':
        semantic_weight, density_weight = 0.7, 0.3
    elif query_type == 'exploratory':
        semantic_weight, density_weight = 0.4, 0.6
    else:
        semantic_weight, density_weight = 0.5, 0.5
    
    multiplicative_score = semantic_scores * density_scores
    weighted_score = semantic_weight * semantic_scores + density_weight * density_scores
    
    return 0.3 * multiplicative_score + 0.7 * weighted_score


# ============================================================
# 分数融合
# ============================================================

def fuse_scores(
    scores1: np.ndarray,
    scores2: np.ndarray,
    weight1: float,
    weight2: float,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    融合两组分数（加权平均）
    
    Args:
        scores1: 第一组分数 (n,)
        scores2: 第二组分数 (n,)
        weight1: 第一组权重
        weight2: 第二组权重
        use_rust: 是否使用 Rust 加速
        
    Returns:
        融合后的分数 (n,)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    scores1 = np.ascontiguousarray(scores1, dtype=np.float32)
    scores2 = np.ascontiguousarray(scores2, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.fuse_scores(scores1, scores2, weight1, weight2)
        except Exception as e:
            logger.warning(f"Rust fuse_scores 失败，回退到 NumPy: {e}")
    
    return weight1 * scores1 + weight2 * scores2


def multiplicative_fuse(
    scores1: np.ndarray,
    scores2: np.ndarray,
    alpha: float = 0.5,
    use_rust: Optional[bool] = None
) -> np.ndarray:
    """
    乘法融合两组分数
    
    Args:
        scores1: 第一组分数 (n,)
        scores2: 第二组分数 (n,)
        alpha: 加权融合的权重
        use_rust: 是否使用 Rust 加速
        
    Returns:
        融合后的分数 (n,)
    """
    if use_rust is None:
        use_rust = RUST_AVAILABLE
    
    scores1 = np.ascontiguousarray(scores1, dtype=np.float32)
    scores2 = np.ascontiguousarray(scores2, dtype=np.float32)
    
    if use_rust:
        try:
            return foxhipporag_rust.multiplicative_fuse(scores1, scores2, alpha)
        except Exception as e:
            logger.warning(f"Rust multiplicative_fuse 失败，回退到 NumPy: {e}")
    
    multiplicative = scores1 * scores2
    weighted = alpha * scores1 + (1.0 - alpha) * scores2
    return 0.3 * multiplicative + 0.7 * weighted



"""
性能优化模块

包含多种性能优化组件：
1. PPR结果缓存
2. 嵌入向量缓存
3. 批量操作优化
4. 内存管理优化
"""

import hashlib
import json
import os
import sqlite3
import threading
import time
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple
import numpy as np
import logging

logger = logging.getLogger(__name__)


class PPRCache:
    """PPR结果缓存
    
    缓存个性化PageRank计算结果，避免重复计算
    """
    
    def __init__(self, cache_dir: str, max_memory_size: int = 100):
        """
        初始化PPR缓存
        
        Args:
            cache_dir: 缓存目录
            max_memory_size: 内存中最大缓存条目数
        """
        os.makedirs(cache_dir, exist_ok=True)
        self.cache_file = os.path.join(cache_dir, "ppr_cache.sqlite")
        self.memory_cache: OrderedDict = OrderedDict()
        self.max_memory_size = max_memory_size
        self.lock = threading.RLock()
        
        self.hits = 0
        self.misses = 0
        
        self._init_db()
    
    def _init_db(self):
        """初始化数据库"""
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute("""
                CREATE TABLE IF NOT EXISTS ppr_cache (
                    query_hash TEXT PRIMARY KEY,
                    doc_ids TEXT,
                    doc_scores TEXT,
                    graph_hash TEXT,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
            """)
            c.execute("CREATE INDEX IF NOT EXISTS idx_graph_hash ON ppr_cache(graph_hash)")
            conn.commit()
            conn.close()
    
    def _compute_hash(self, reset_prob: np.ndarray, damping: float, graph_hash: str) -> str:
        """计算缓存键的哈希值"""
        content = f"{graph_hash}:{damping}:{np.array2string(reset_prob, precision=6)}"
        return hashlib.md5(content.encode()).hexdigest()
    
    def get(self, reset_prob: np.ndarray, damping: float, graph_hash: str) -> Optional[Tuple[np.ndarray, np.ndarray]]:
        """
        获取缓存的PPR结果
        
        Args:
            reset_prob: 重置概率数组
            damping: 阻尼因子
            graph_hash: 图的哈希值
            
        Returns:
            缓存的(doc_ids, doc_scores)或None
        """
        cache_key = self._compute_hash(reset_prob, damping, graph_hash)
        
        with self.lock:
            if cache_key in self.memory_cache:
                self.memory_cache.move_to_end(cache_key)
                self.hits += 1
                return self.memory_cache[cache_key]
        
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute(
                "SELECT doc_ids, doc_scores FROM ppr_cache WHERE query_hash = ? AND graph_hash = ?",
                (cache_key, graph_hash)
            )
            row = c.fetchone()
            conn.close()
            
            if row:
                doc_ids = np.array(json.loads(row[0]))
                doc_scores = np.array(json.loads(row[1]))
                
                with self.lock:
                    if len(self.memory_cache) >= self.max_memory_size:
                        self.memory_cache.popitem(last=False)
                    self.memory_cache[cache_key] = (doc_ids, doc_scores)
                
                self.hits += 1
                return doc_ids, doc_scores
        
        self.misses += 1
        return None
    
    def put(self, reset_prob: np.ndarray, damping: float, graph_hash: str,
            doc_ids: np.ndarray, doc_scores: np.ndarray) -> None:
        """
        存储PPR结果到缓存
        
        Args:
            reset_prob: 重置概率数组
            damping: 阻尼因子
            graph_hash: 图的哈希值
            doc_ids: 文档ID数组
            doc_scores: 文档得分数组
        """
        cache_key = self._compute_hash(reset_prob, damping, graph_hash)
        
        with self.lock:
            if len(self.memory_cache) >= self.max_memory_size:
                self.memory_cache.popitem(last=False)
            self.memory_cache[cache_key] = (doc_ids.copy(), doc_scores.copy())
        
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute(
                """INSERT OR REPLACE INTO ppr_cache 
                   (query_hash, doc_ids, doc_scores, graph_hash) 
                   VALUES (?, ?, ?, ?)""",
                (cache_key, json.dumps(doc_ids.tolist()), json.dumps(doc_scores.tolist()), graph_hash)
            )
            conn.commit()
            conn.close()
    
    def invalidate(self, graph_hash: str = None):
        """
        使缓存失效
        
        Args:
            graph_hash: 如果提供，只使该图的缓存失效
        """
        with self.lock:
            self.memory_cache.clear()
        
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            if graph_hash:
                c.execute("DELETE FROM ppr_cache WHERE graph_hash = ?", (graph_hash,))
            else:
                c.execute("DELETE FROM ppr_cache")
            conn.commit()
            conn.close()
    
    def get_stats(self) -> Dict[str, Any]:
        """获取缓存统计信息"""
        with self.lock:
            total = self.hits + self.misses
            hit_rate = self.hits / total if total > 0 else 0
            return {
                'memory_size': len(self.memory_cache),
                'max_memory_size': self.max_memory_size,
                'hits': self.hits,
                'misses': self.misses,
                'hit_rate': hit_rate
            }


class EmbeddingCache:
    """嵌入向量缓存
    
    缓存文本的嵌入向量，避免重复计算
    """
    
    def __init__(self, cache_dir: str, max_memory_size: int = 1000):
        """
        初始化嵌入缓存
        
        Args:
            cache_dir: 缓存目录
            max_memory_size: 内存中最大缓存条目数
        """
        os.makedirs(cache_dir, exist_ok=True)
        self.cache_file = os.path.join(cache_dir, "embedding_cache.sqlite")
        self.memory_cache: OrderedDict = OrderedDict()
        self.max_memory_size = max_memory_size
        self.lock = threading.RLock()
        
        self.hits = 0
        self.misses = 0
        
        self._init_db()
    
    def _init_db(self):
        """初始化数据库"""
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute("""
                CREATE TABLE IF NOT EXISTS embedding_cache (
                    text_hash TEXT PRIMARY KEY,
                    embedding TEXT,
                    model_name TEXT,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
            """)
            c.execute("CREATE INDEX IF NOT EXISTS idx_model_name ON embedding_cache(model_name)")
            conn.commit()
            conn.close()
    
    def _compute_hash(self, text: str, model_name: str) -> str:
        """计算文本哈希"""
        content = f"{model_name}:{text}"
        return hashlib.md5(content.encode()).hexdigest()
    
    def get(self, text: str, model_name: str) -> Optional[np.ndarray]:
        """获取缓存的嵌入向量"""
        cache_key = self._compute_hash(text, model_name)
        
        with self.lock:
            if cache_key in self.memory_cache:
                self.memory_cache.move_to_end(cache_key)
                self.hits += 1
                return self.memory_cache[cache_key]
        
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute(
                "SELECT embedding FROM embedding_cache WHERE text_hash = ? AND model_name = ?",
                (cache_key, model_name)
            )
            row = c.fetchone()
            conn.close()
            
            if row:
                embedding = np.array(json.loads(row[0]))
                
                with self.lock:
                    if len(self.memory_cache) >= self.max_memory_size:
                        self.memory_cache.popitem(last=False)
                    self.memory_cache[cache_key] = embedding
                
                self.hits += 1
                return embedding
        
        self.misses += 1
        return None
    
    def get_batch(self, texts: List[str], model_name: str) -> Dict[int, np.ndarray]:
        """
        批量获取缓存的嵌入向量
        
        Args:
            texts: 文本列表
            model_name: 模型名称
            
        Returns:
            索引到嵌入向量的映射（只包含缓存命中的）
        """
        results = {}
        uncached_indices = []
        
        with self.lock:
            for i, text in enumerate(texts):
                cache_key = self._compute_hash(text, model_name)
                if cache_key in self.memory_cache:
                    self.memory_cache.move_to_end(cache_key)
                    results[i] = self.memory_cache[cache_key]
                    self.hits += 1
                else:
                    uncached_indices.append(i)
        
        if uncached_indices:
            uncached_texts = [texts[i] for i in uncached_indices]
            cache_keys = [self._compute_hash(t, model_name) for t in uncached_texts]
            
            with self.lock:
                conn = sqlite3.connect(self.cache_file)
                c = conn.cursor()
                placeholders = ','.join(['?' for _ in cache_keys])
                c.execute(
                    f"SELECT text_hash, embedding FROM embedding_cache WHERE text_hash IN ({placeholders}) AND model_name = ?",
                    cache_keys + [model_name]
                )
                rows = c.fetchall()
                conn.close()
                
                db_results = {row[0]: np.array(json.loads(row[1])) for row in rows}
                
                for i, cache_key in zip(uncached_indices, cache_keys):
                    if cache_key in db_results:
                        embedding = db_results[cache_key]
                        results[i] = embedding
                        
                        with self.lock:
                            if len(self.memory_cache) >= self.max_memory_size:
                                self.memory_cache.popitem(last=False)
                            self.memory_cache[cache_key] = embedding
                        
                        self.hits += 1
                    else:
                        self.misses += 1
        
        return results
    
    def put(self, text: str, model_name: str, embedding: np.ndarray) -> None:
        """存储嵌入向量到缓存"""
        cache_key = self._compute_hash(text, model_name)
        
        with self.lock:
            if len(self.memory_cache) >= self.max_memory_size:
                self.memory_cache.popitem(last=False)
            self.memory_cache[cache_key] = embedding.copy()
        
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute(
                "INSERT OR REPLACE INTO embedding_cache (text_hash, embedding, model_name) VALUES (?, ?, ?)",
                (cache_key, json.dumps(embedding.tolist()), model_name)
            )
            conn.commit()
            conn.close()
    
    def put_batch(self, texts: List[str], model_name: str, embeddings: np.ndarray) -> None:
        """批量存储嵌入向量"""
        with self.lock:
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            
            for text, embedding in zip(texts, embeddings):
                cache_key = self._compute_hash(text, model_name)
                
                if len(self.memory_cache) >= self.max_memory_size:
                    self.memory_cache.popitem(last=False)
                self.memory_cache[cache_key] = embedding.copy()
                
                c.execute(
                    "INSERT OR REPLACE INTO embedding_cache (text_hash, embedding, model_name) VALUES (?, ?, ?)",
                    (cache_key, json.dumps(embedding.tolist()), model_name)
                )
            
            conn.commit()
            conn.close()
    
    def get_stats(self) -> Dict[str, Any]:
        """获取缓存统计信息"""
        with self.lock:
            total = self.hits + self.misses
            hit_rate = self.hits / total if total > 0 else 0
            return {
                'memory_size': len(self.memory_cache),
                'max_memory_size': self.max_memory_size,
                'hits': self.hits,
                'misses': self.misses,
                'hit_rate': hit_rate
            }


class BatchProcessor:
    """批量处理器
    
    优化批量操作的性能
    """
    
    def __init__(self, batch_size: int = 32, max_workers: int = 8):
        """
        初始化批量处理器
        
        Args:
            batch_size: 批处理大小
            max_workers: 最大工作线程数
        """
        self.batch_size = batch_size
        self.max_workers = max_workers
    
    def process_in_batches(self, items: List[Any], process_fn, desc: str = "Processing") -> List[Any]:
        """
        分批处理项目
        
        Args:
            items: 待处理的项目列表
            process_fn: 处理函数
            desc: 进度条描述
            
        Returns:
            处理结果列表
        """
        from concurrent.futures import ThreadPoolExecutor, as_completed
        from tqdm import tqdm
        
        if not items:
            return []
        
        if len(items) <= self.batch_size:
            return [process_fn(item) for item in tqdm(items, desc=desc)]
        
        batches = [items[i:i + self.batch_size] for i in range(0, len(items), self.batch_size)]
        
        results = [None] * len(items)
        
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            future_to_batch = {
                executor.submit(self._process_batch, batch, process_fn, start_idx): (batch, start_idx)
                for start_idx, batch in enumerate(batches)
            }
            
            for future in tqdm(as_completed(future_to_batch), total=len(future_to_batch), desc=desc):
                batch, start_idx = future_to_batch[future]
                try:
                    batch_results = future.result()
                    for i, result in enumerate(batch_results):
                        results[start_idx * self.batch_size + i] = result
                except Exception as e:
                    logger.error(f"Batch processing failed: {e}")
        
        return results
    
    def _process_batch(self, batch: List[Any], process_fn, start_idx: int) -> List[Any]:
        """处理单个批次"""
        return [process_fn(item) for item in batch]


class MemoryManager:
    """内存管理器
    
    监控和管理内存使用
    """
    
    def __init__(self, warning_threshold: float = 0.8, critical_threshold: float = 0.9):
        """
        初始化内存管理器
        
        Args:
            warning_threshold: 警告阈值（内存使用比例）
            critical_threshold: 临界阈值（内存使用比例）
        """
        self.warning_threshold = warning_threshold
        self.critical_threshold = critical_threshold
        self._callbacks = []
    
    def add_cleanup_callback(self, callback):
        """添加清理回调函数"""
        self._callbacks.append(callback)
    
    def check_memory(self) -> Tuple[float, str]:
        """
        检查内存使用情况
        
        Returns:
            (内存使用比例, 状态)
        """
        try:
            import psutil
            memory = psutil.virtual_memory()
            usage = memory.percent / 100.0
            
            if usage >= self.critical_threshold:
                status = "critical"
                self._trigger_cleanup()
            elif usage >= self.warning_threshold:
                status = "warning"
            else:
                status = "normal"
            
            return usage, status
        except ImportError:
            return 0.0, "unknown"
    
    def _trigger_cleanup(self):
        """触发清理操作"""
        logger.warning("Memory usage critical, triggering cleanup...")
        for callback in self._callbacks:
            try:
                callback()
            except Exception as e:
                logger.error(f"Cleanup callback failed: {e}")
    
    def get_memory_info(self) -> Dict[str, Any]:
        """获取内存信息"""
        try:
            import psutil
            memory = psutil.virtual_memory()
            return {
                'total': memory.total,
                'available': memory.available,
                'used': memory.used,
                'percent': memory.percent,
                'status': self.check_memory()[1]
            }
        except ImportError:
            return {'status': 'psutil not available'}


class PerformanceMonitor:
    """性能监控器
    
    监控和记录性能指标
    """
    
    def __init__(self):
        """初始化性能监控器"""
        self.metrics = {}
        self.lock = threading.Lock()
    
    def record(self, name: str, value: float):
        """记录性能指标"""
        with self.lock:
            if name not in self.metrics:
                self.metrics[name] = []
            self.metrics[name].append({
                'value': value,
                'timestamp': time.time()
            })
    
    def record_time(self, name: str):
        """记录时间的上下文管理器"""
        class Timer:
            def __init__(self, monitor, metric_name):
                self.monitor = monitor
                self.metric_name = metric_name
                self.start_time = None
            
            def __enter__(self):
                self.start_time = time.time()
                return self
            
            def __exit__(self, *args):
                elapsed = time.time() - self.start_time
                self.monitor.record(self.metric_name, elapsed)
        
        return Timer(self, name)
    
    def get_stats(self, name: str) -> Dict[str, float]:
        """获取指标统计信息"""
        with self.lock:
            if name not in self.metrics:
                return {}
            
            values = [m['value'] for m in self.metrics[name]]
            if not values:
                return {}
            
            return {
                'count': len(values),
                'mean': np.mean(values),
                'std': np.std(values),
                'min': np.min(values),
                'max': np.max(values),
                'total': np.sum(values)
            }
    
    def get_all_stats(self) -> Dict[str, Dict[str, float]]:
        """获取所有指标的统计信息"""
        with self.lock:
            return {name: self.get_stats(name) for name in self.metrics}
    
    def clear(self):
        """清除所有指标"""
        with self.lock:
            self.metrics.clear()


global_performance_monitor = PerformanceMonitor()



def get_query_instruction(linking_method):
    instructions = {
        'ner_to_node': 'Given a phrase, retrieve synonymous or relevant phrases that best match this phrase.',
        'query_to_node': 'Given a question, retrieve relevant phrases that are mentioned in this question.',
        'query_to_fact': 'Given a question, retrieve relevant triplet facts that matches this question.',
        'query_to_sentence': 'Given a question, retrieve relevant sentences that best answer the question.',
        'query_to_passage': 'Given a question, retrieve relevant documents that best answer the question.',
    }
    default_instruction = 'Given a question, retrieve relevant documents that best answer the question.'
    return instructions.get(linking_method, default_instruction)
    


from copy import deepcopy
from typing import List, Optional

import numpy as np
import torch
from tqdm import tqdm
from transformers import AutoModel

from ..utils.config_utils import BaseConfig
from ..utils.logging_utils import get_logger
from .base import BaseEmbeddingModel, EmbeddingConfig

logger = get_logger(__name__)


class NVEmbedV2EmbeddingModel(BaseEmbeddingModel):

    def __init__(self, global_config: Optional[BaseConfig] = None, embedding_model_name: Optional[str] = None) -> None:
        super().__init__(global_config=global_config)

        if embedding_model_name is not None:
            self.embedding_model_name = embedding_model_name
            logger.debug(f"Overriding {self.__class__.__name__}'s embedding_model_name with: {self.embedding_model_name}")

        self._init_embedding_config()

        # Initializing the embedding model
        logger.debug(f"Initializing {self.__class__.__name__}'s embedding model with params: {self.embedding_config.model_init_params}")

        self.embedding_model = AutoModel.from_pretrained(**self.embedding_config.model_init_params)
        self.embedding_dim = self.embedding_model.config.hidden_size

    def _init_embedding_config(self) -> None:
        """
        Extract embedding model-specific parameters to init the EmbeddingConfig.
        
        Returns:
            None
        """

        config_dict = {
            "embedding_model_name": self.embedding_model_name,
            "norm": self.global_config.embedding_return_as_normalized,
            # "max_seq_length": self.global_config.embedding_max_seq_len,
            "model_init_params": {
                # "model_name_or_path": self.embedding_model_name2mode_name_or_path[self.embedding_model_name],
                "pretrained_model_name_or_path": self.embedding_model_name,
                "trust_remote_code": True,
                'device_map': "auto",  # added this line to use multiple GPUs
                "torch_dtype": self.global_config.embedding_model_dtype,
                # **kwargs
            },
            "encode_params": {
                "max_length": self.global_config.embedding_max_seq_len,  # 32768 from official example,
                "instruction": "",
                "batch_size": self.global_config.embedding_batch_size,
                "num_workers": 32
            },
        }

        self.embedding_config = EmbeddingConfig.from_dict(config_dict=config_dict)
        logger.debug(f"Init {self.__class__.__name__}'s embedding_config: {self.embedding_config}")

    # def _add_eos(self, texts: List[str]) -> List[str]:
    #     # Adds EOS token to each text
    #     return [text + self.embedding_model.tokenizer.eos_token for text in texts]

    def batch_encode(self, texts: List[str], **kwargs) -> None:
        if isinstance(texts, str):
            texts = [texts]

        params = deepcopy(self.embedding_config.encode_params)
        if kwargs:
            params.update(kwargs)

        if "instruction" in kwargs:
            if kwargs["instruction"] != '':
                params["instruction"] = f"Instruct: {kwargs['instruction']}\nQuery: "
            # del params["instruction"]

        batch_size = params.pop("batch_size", 16)

        logger.debug(f"Calling {self.__class__.__name__} with:\n{params}")
        if len(texts) <= batch_size:
            params["prompts"] = texts  # self._add_eos(texts=texts)
            results = self.embedding_model.encode(**params)
        else:
            pbar = tqdm(total=len(texts), desc="Batch Encoding")
            results = []
            for i in range(0, len(texts), batch_size):
                params["prompts"] = texts[i:i + batch_size]
                results.append(self.embedding_model.encode(**params))
                pbar.update(batch_size)
            pbar.close()
            results = torch.cat(results, dim=0)

        if isinstance(results, torch.Tensor):
            results = results.cpu()
            results = results.numpy()
        if self.embedding_config.norm:
            results = (results.T / np.linalg.norm(results, axis=1)).T

        return results



from typing import List
import json

import boto3
from botocore.exceptions import ClientError
import numpy as np
from tqdm import tqdm

from .base import BaseEmbeddingModel
from ..utils.config_utils import BaseConfig
from ..prompts.linking import get_query_instruction


class CohereEmbeddingModel(BaseEmbeddingModel):
    """
    To select this implementation you can initialise HippoRAG with:
        embedding_model_name="cohere.embed-english-v3"
    """
    def __init__(self, global_config:BaseConfig, embedding_model_name:str) -> None:
        super().__init__(global_config=global_config)

        self.model_id = embedding_model_name
        self.embedding_type = 'float'
        self.batch_size = 64

        self.bedrock_runtime = boto3.client(service_name='bedrock-runtime')

        self.search_query_instr = set([
            get_query_instruction('query_to_fact'),
            get_query_instruction('query_to_passage')
        ])

    def encode(self, texts: List[str], input_type) -> None:
        request = {
             'texts': texts,
             'input_type': input_type,
             'embedding_types': [self.embedding_type]
        }
        try:
            response = self.bedrock_runtime.invoke_model(
                body=json.dumps(request),
                modelId=self.model_id,
                accept='*/*',
                contentType='application/json'
            )
        except ClientError as err:
            raise Exception(f"A client error occurred: {err.response['Error']['Message']}")
        
        response = json.loads(response.get('body').read())
        return np.array(response['embeddings'][self.embedding_type])

    def batch_encode(self, texts: List[str], **kwargs) -> None:
        input_type = 'search_query' if (kwargs.get("instruction") in self.search_query_instr) else 'search_document'

        if len(texts) < self.batch_size:
            return self.encode(texts, input_type)
        
        results = []
        batch_indexes = list(range(0, len(texts), self.batch_size))
        for i in tqdm(batch_indexes, desc="Batch Encoding"):
            results.append(self.encode(texts[i:i + self.batch_size], input_type))
        return np.concatenate(results)



from typing import List
import torch.cuda
import outlines.generate as generate
import outlines.models as models
from transformers import PreTrainedTokenizer, AutoModelForCausalLM, AutoTokenizer

from .base import LLMConfig
from ..utils.llm_utils import TextChatMessage, get_pydantic_model
from ..utils.logging_utils import get_logger

logger = get_logger(__name__)

def convert_text_chat_messages_to_strings(messages: List[TextChatMessage], tokenizer: PreTrainedTokenizer, add_assistant_header=True) -> List[str]:
    return tokenizer.apply_chat_template(conversation=messages, tokenize=False)

def convert_text_chat_messages_to_input_string(messages: List[TextChatMessage], tokenizer: PreTrainedTokenizer, add_assistant_header=True) -> str:
    prompt = tokenizer.apply_chat_template(
        conversation=messages,
        chat_template=None,
        tokenize=False,
        add_generation_prompt=True,
        continue_final_message=False,
        tools=None,
        documents=None,
    )
    return prompt

class TransformersOffline:

    def _init_llm_config(self) -> None:
        self.llm_config = LLMConfig()

    def __init__(self, global_config, cache_dir=None, cache_filename=None, max_model_len=4096, **kwargs):
        model_name = kwargs.get('model_name', global_config.llm_name)
        if model_name is None:
            model_name = 'meta-llama/Llama-3.1-8B-Instruct'
        import os
        self.model_name = model_name
        self.model = AutoModelForCausalLM.from_pretrained(model_name, device_map='auto', torch_dtype = torch.bfloat16)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        
        if cache_filename is None:
            cache_filename = f'{model_name.replace("/", "_")}_cache.sqlite'
        if cache_dir is None:
            cache_dir = os.path.join(global_config.save_dir, "llm_cache")
        self.cache_file_name = os.path.join(cache_dir, cache_filename)
    
    def infer(self, messages: List[TextChatMessage], max_tokens=2048):
        logger.info(f"Calling Transformers offline, # of messages {len(messages)}")
        messages_list = [messages]
        prompt_text = convert_text_chat_messages_to_input_string(messages_list, self.tokenizer)
        input_ids = self.tokenizer.encode(prompt_text)
        transformers_output = self.model.generate(input_ids, max_new_tokens=max_tokens)
        response = transformers_output[0]['generated_text']
        prompt_tokens = len(self.tokenizer.encode(prompt_text))
        completion_tokens = len(self.tokenizer.encode(response))
        metadata = {
            "prompt_tokens": prompt_tokens,
            "completion_tokens": completion_tokens
        }
        return response, metadata

    def batch_infer(self, messages_list: List[List[TextChatMessage]], max_tokens=2048, json_template=None):
        if len(messages_list) > 1:
            logger.info(f"Calling Transformers offline, # of messages {len(messages_list)}, using batchsize = 4")

        all_prompt_texts = [convert_text_chat_messages_to_input_string(messages, self.tokenizer) for messages in messages_list]

        if json_template is not None:
            guided_json=get_pydantic_model(json_template)
            outlines_model = models.Transformers(self.model, self.tokenizer)
            generator = generate.json(outlines_model, guided_json)
            transformers_outputs = []
            for i in range(0, len(all_prompt_texts), 4):
                transformers_output = generator(all_prompt_texts[i:i+4], max_tokens=max_tokens)
                transformers_outputs.extend(transformers_output)
        else:
            transformers_outputs = []
            for i in range(0, len(all_prompt_texts), 4):
                transformers_output = self.model.generate(all_prompt_texts[i:i+4], max_tokens=max_tokens)
                transformers_outputs.extend(transformers_output)
        all_responses = [completion.model_dump_json() for completion in transformers_outputs]
        all_prompt_tokens = [len(self.tokenizer.encode(prompt)) for prompt in all_prompt_texts]
        all_completion_tokens = [len(self.tokenizer.encode(response)) for response in all_responses]

        metadata = {
            "prompt_tokens": sum(all_prompt_tokens),
            "completion_tokens": sum(all_completion_tokens),
            "num_request": len(messages_list)
        }
        return all_responses, metadata



"""
向量数据库集成模块

支持的向量数据库后端：
- Parquet: 默认的本地文件存储（基于 Parquet 格式）
- FAISS: Facebook AI Similarity Search，高效的本地向量检索
- Milvus: 开源分布式向量数据库
- Chroma: 轻量级嵌入式向量数据库
"""

from .base import BaseVectorDB
from .parquet_db import ParquetVectorDB
from .faiss_db import FAISSVectorDB

# 可选导入，仅在安装了相应依赖时可用
try:
    from .milvus_db import MilvusVectorDB
except ImportError:
    MilvusVectorDB = None

try:
    from .chroma_db import ChromaVectorDB
except ImportError:
    ChromaVectorDB = None


def get_vector_db(backend: str, **kwargs) -> BaseVectorDB:
    """
    工厂函数：根据配置获取向量数据库实例
    
    Args:
        backend: 向量数据库后端类型 ('parquet', 'faiss', 'milvus', 'chroma')
        **kwargs: 传递给具体实现的参数
        
    Returns:
        BaseVectorDB: 向量数据库实例
        
    Raises:
        ValueError: 不支持的后端类型
        ImportError: 未安装必要的依赖
    """
    backend = backend.lower()
    
    if backend == 'parquet':
        return ParquetVectorDB(**kwargs)
    elif backend == 'faiss':
        return FAISSVectorDB(**kwargs)
    elif backend == 'milvus':
        if MilvusVectorDB is None:
            raise ImportError(
                "Milvus 支持未安装。请运行: pip install pymilvus>=2.3.0"
            )
        return MilvusVectorDB(**kwargs)
    elif backend == 'chroma':
        if ChromaVectorDB is None:
            raise ImportError(
                "Chroma 支持未安装。请运行: pip install chromadb>=0.4.0"
            )
        return ChromaVectorDB(**kwargs)
    else:
        raise ValueError(
            f"不支持的向量数据库后端: {backend}。"
            f"支持的后端: parquet, faiss, milvus, chroma"
        )


__all__ = [
    'BaseVectorDB',
    'ParquetVectorDB', 
    'FAISSVectorDB',
    'MilvusVectorDB',
    'ChromaVectorDB',
    'get_vector_db',
]



from typing import (
    List,
    Optional
)
import torch
import numpy as np
from copy import deepcopy
from gritlm import GritLM

from ..utils.config_utils import BaseConfig
from ..utils.logging_utils import get_logger
from ..utils.llm_utils import TextChatMessage

from .base import BaseEmbeddingModel, EmbeddingConfig


logger = get_logger(__name__)


class GritLMEmbeddingModel(BaseEmbeddingModel):

    def __init__(self, global_config: Optional[BaseConfig] = None, embedding_model_name: Optional[str] = None) -> None:
        super().__init__(global_config=global_config)
        
        if embedding_model_name is not None:
            self.embedding_model_name = embedding_model_name
            logger.debug(f"Overriding {self.__class__.__name__}'s embedding_model_name with: {self.embedding_model_name}")
        
        self._init_embedding_config()

        # Initializing the embedding model
        logger.debug(f"Initializing {self.__class__.__name__}'s embedding model with params: {self.embedding_config.model_init_params}")
        self.embedding_model = GritLM(**self.embedding_config.model_init_params)
        self.embedding_dim = self.embedding_model.model.config.hidden_size

        self.device = self.embedding_model.device

        
    def _init_embedding_config(self) -> None:
        """
        Extract embedding model-specific parameters to init the EmbeddingConfig.
        
        Returns:
            None
        """

        config_dict = {
            "embedding_model_name": self.embedding_model_name,
            "norm": self.global_config.embedding_return_as_normalized,
            "model_init_params": {
                "model_name_or_path": self.embedding_model_name,
                "torch_dtype": self.global_config.embedding_model_dtype,
                "device_map": "auto", # added this line to use multiple GPUs
                # **kwargs
            },
            "encode_params": {
                "batch_size": self.global_config.embedding_batch_size,
            },
            "generate_params": {
            }
        }

        self.embedding_config = EmbeddingConfig.from_dict(config_dict=config_dict)
        logger.debug(f"Init {self.__class__.__name__}'s embedding_config: {self.embedding_config}")
    
    
    def _get_formated_instruction(self, instruction: str) -> str:
        return "<|user|>\n" + instruction + "\n<|embed|>\n" if instruction else "<|embed|>\n"
    
    
    def batch_encode(self, texts: List[str], **kwargs) -> None:
        if isinstance(texts, str):
            texts = [texts]

        params = deepcopy(self.embedding_config.encode_params)
        if kwargs:
            params.update(kwargs)
        if "instruction" in kwargs:
            params["instruction"] = self._get_formated_instruction(params["instruction"])
        params["sentences"] = texts
        
        
        logger.debug(f"Calling {self.__class__.__name__} with:\n{params}")
        results = self.embedding_model.encode(**params)

        if isinstance(results, torch.Tensor):
            results = results.cpu()
            results = results.numpy()
        if self.embedding_config.norm:
            results = (results.T / np.linalg.norm(results, axis=1)).T
        
        return results
        
    
    def batch_generate(self, chat: List[TextChatMessage],) -> None:
        pass
    
    
    


import json
import re
import pydantic
from string import Template
from typing import (
    Optional,
    Union,
    List,
    TypedDict,
    Tuple,
    Dict,
    Type
)

from openai import (
    APIConnectionError,
    RateLimitError,
    Timeout
)
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception_type,
)

from .config_utils import BaseConfig


class TextChatMessage(TypedDict):
    """Representation of a single text-based chat message in the chat history."""
    role: str  # Either "system", "user", or "assistant"
    content: Union[str, Template]  # The text content of the message (could also be a string.Template instance)


def convert_format_to_template(original_string: str, placeholder_mapping: Optional[dict] = None, static_values: Optional[dict] = None) -> str:
    """
    Converts a .format() style string to a Template-style string.

    Args:
        original_string (str): The original string using .format() placeholders.
        placeholder_mapping (dict, optional): Mapping from original placeholder names to new placeholder names.
        static_values (dict, optional): Mapping from original placeholders to static values to be replaced in the new template.

    Returns:
        str: The converted string in Template-style format.
    """
    # Initialize mappings
    placeholder_mapping = placeholder_mapping or {}
    static_values = static_values or {}

    # Regular expression to find .format() style placeholders
    placeholder_pattern = re.compile(r'\{(\w+)\}')

    # Substitute placeholders in the string
    def replace_placeholder(match):
        original_placeholder = match.group(1)

        # If the placeholder is in static_values, substitute its value directly
        if original_placeholder in static_values:
            return str(static_values[original_placeholder])

        # Otherwise, rename the placeholder if needed, or keep it as is
        new_placeholder = placeholder_mapping.get(original_placeholder, original_placeholder)
        return f'${{{new_placeholder}}}'

    # Replace all placeholders
    template_string = placeholder_pattern.sub(replace_placeholder, original_string)

    return template_string


def safe_unicode_decode(content: Union[bytes, str]) -> str:
    """
    Safely decodes a string or byte sequence containing Unicode escape sequences into a properly decoded Unicode string.

    Args:
        content (Union[bytes, str]): The input data, either as a byte sequence or a string, which may contain Unicode escape sequences.

    Returns:
        str: The decoded Unicode string with escape sequences replaced by their corresponding characters.
    
    Raises:
        AttributeError: If the input is neither bytes nor a string.
    """
    # Regular expression to find all Unicode escape sequences of the form \uXXXX
    unicode_escape_pattern = re.compile(r"\\u([0-9a-fA-F]{4})")

    # Function to replace the Unicode escape with the actual character
    def replace_unicode_escape(match: re.Match) -> str:
        # Convert the matched hexadecimal value into the actual Unicode character
        return chr(int(match.group(1), 16))

    # If content is bytes, decode it to a UTF-8 string
    if isinstance(content, bytes):
        content = content.decode("utf-8")
    elif not isinstance(content, str):
        raise AttributeError("Input must be of type bytes or str.")

    # Perform the substitution
    decoded_content = unicode_escape_pattern.sub(replace_unicode_escape, content)

    return decoded_content


def dynamic_retry(experiment_config: BaseConfig):
    """
    Factory function to create a retry decorator with dynamic parameters.

    Args:
        experiment_config (BaseConfig): Configuration containing stop and wait parameters. Expected to use the global config for running all experiments. 

    Returns:
        Callable: A retry decorator with dynamically set parameters.
    """
    stop = stop_after_attempt(experiment_config.async_max_retry_attempts)
    wait = wait_exponential(
        multiplier=experiment_config.async_retry_wait_exp_multiplier,
        min=experiment_config.async_retry_min_wait_exp_time,
        max=experiment_config.async_retry_max_wait_exp_time,
    )

    return retry(
        stop=stop,
        wait=wait,
        retry=retry_if_exception_type((RateLimitError, APIConnectionError, Timeout)),
    )


# def fix_broken_generated_json(json_str: str) -> str:
#     """
#     Fixes a malformed JSON string by:
#     - Removing the last comma and any trailing content.
#     - Appending a closing bracket `]` and brace `}` to properly terminate the JSON.

#     Args:
#         json_str (str): The malformed JSON string to be fixed.

#     Returns:
#         str: The corrected JSON string.
#     """
#     last_comma_index = json_str.rfind(',')
#     if last_comma_index != -1:
#         json_str = json_str[:last_comma_index]

#     processed_string = json_str + ']\n}'
#     return processed_string


def fix_broken_generated_json(json_str: str) -> str:
    """
    Fixes a malformed JSON string by:
    - Removing the last comma and any trailing content.
    - Iterating over the JSON string once to determine and fix unclosed braces or brackets.
    - Ensuring braces and brackets inside string literals are not considered.

    If the original json_str string can be successfully loaded by json.loads(), will directly return it without any modification.
    
    Args:
        json_str (str): The malformed JSON string to be fixed.

    Returns:
        str: The corrected JSON string.
    """

    def find_unclosed(json_str):
        """
        Identifies the unclosed braces and brackets in the JSON string.

        Args:
            json_str (str): The JSON string to analyze.

        Returns:
            list: A list of unclosed elements in the order they were opened.
        """
        unclosed = []
        inside_string = False
        escape_next = False

        for char in json_str:
            if inside_string:
                if escape_next:
                    escape_next = False
                elif char == '\\':
                    escape_next = True
                elif char == '"':
                    inside_string = False
            else:
                if char == '"':
                    inside_string = True
                elif char in '{[':
                    unclosed.append(char)
                elif char in '}]':
                    if unclosed and ((char == '}' and unclosed[-1] == '{') or (char == ']' and unclosed[-1] == '[')):
                        unclosed.pop()

        return unclosed

    try:
        # Try to load the JSON to see if it is valid
        json.loads(json_str)
        return json_str  # Return as-is if valid
    except json.JSONDecodeError:
        pass

    # Step 1: Remove trailing content after the last comma.
    last_comma_index = json_str.rfind(',')
    if last_comma_index != -1:
        json_str = json_str[:last_comma_index]

    # Step 2: Identify unclosed braces and brackets.
    unclosed_elements = find_unclosed(json_str)

    # Step 3: Append the necessary closing elements in reverse order of opening.
    closing_map = {'{': '}', '[': ']'}
    for open_char in reversed(unclosed_elements):
        json_str += closing_map[open_char]

    return json_str


def filter_invalid_triples(triples: List[List[str]]) -> List[List[str]]:
    """
    Filters out invalid and duplicate triples from a list of triples.

    A valid triple meets the following criteria:
    1. It contains exactly three elements.
    2. It is unique within the list (no duplicates in the output).

    The function ensures:
    - Each valid triple is converted to a list of strings.
    - The order of unique, valid triples is preserved.
    - Do not apply any text preprocessing techniques or rules within this function.
    
    Args:
        triples (List[List[str]]): 
            A list of triples (each a list of strings or elements that can be converted to strings).

    Returns:
        List[List[str]]: 
            A list of unique, valid triples, each represented as a list of strings.
    """
    unique_triples = set()
    valid_triples = []

    for triple in triples:
        if len(triple) != 3:
            continue  # Skip triples that do not have exactly 3 elements

        valid_triple = [str(item) for item in triple]
        if tuple(valid_triple) not in unique_triples:
            unique_triples.add(tuple(valid_triple))
            valid_triples.append(valid_triple)

    return valid_triples


PROMPT_JSON_TEMPLATE = {
    "ner": {
        "type": "object",
        "properties": {
            "named_entities": {
                "type": "array",
                "items": {
                    "type": "string"
                },
                "minItems": 0
            }
        },
        "required": ["named_entities"]
    },
    "triples": {
        "type": "object",
        "properties": {
            "triples": {
                "type": "array",
                "items": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "maxItems": 3,
                    "minItems": 3,
                },
                "minItems": 0
            }
        },
        "required": ["triples"]
    },
    "fact": {
        "type": "object",
        "properties": {
            "fact": {
                "type": "array",
                "items": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "maxItems": 3,
                    "minItems": 3,
                },
                "minItems": 0
            }
        },
        "required": ["fact"]
    },
    "json": {
        "type": "object",
    },
    "qa_cot": {
        "type": "object",
        "required": ["Thought", "Answer"],
        "properties": {
            "Thought": {
                "type": "string",
                "minLength": 1,
                "maxLength": 2000
            },
            "Answer": {
                "type": "string",
                "minLength": 1,
                "maxLength": 200
            },
        },
    }
}


def num_tokens_by_tiktoken(text: str):
    import tiktoken
    enc = tiktoken.encoding_for_model("gpt-3.5-turbo")

    return len(enc.encode(text))


class NerModel(pydantic.BaseModel):
    """
    For NER, the structure is:
    schema:
    {
        "type": "object",
        "properties": { "named_entities": { "type": "array", "items": { "type": "string" } } },
        "required": ["named_entities"]
    }
    """
    named_entities: List[str]

class TriplesModel(pydantic.BaseModel):
    """
    For triples extraction, the structure is:
    schema:
    {
        "type": "object",
        "properties": { "triples": { "type": "array", "items": { "type": "array", "items": { "type": "string" }, "maxItems": 3, "minItems": 3 } } },
        "required": ["triples"]
    }
    """
    triples: List[Tuple[str, str, str]]

class FactModel(pydantic.BaseModel):
    """
    For fact extraction, the structure is:
    schema:
    {
        "type": "object",
        "properties": { "fact": { "type": "array", "items": { "type": "array", "items": { "type": "string" }, "maxItems": 3, "minItems": 3 } } },
        "required": ["fact"]
    }
    """
    fact: List[Tuple[str, str, str]]

class ArbitraryJsonModel(pydantic.BaseModel):
    """
    For generating arbitrary JSON objects.
    In Pydantic V2, by setting extra='allow', this model can accept any fields that are not explicitly defined in the model.
    This fully matches the loose schema definition of { "type": "object" }.
    schema:
    {
        "type": "object"
    }
    """
    model_config = pydantic.ConfigDict(extra='allow')

class QaCotModel(pydantic.BaseModel):
    """
    For Chain-of-Thought QA, the structure is:
    schema:
    {
        "type": "object",
        "required": ["Thought", "Answer"],
        "properties": {
            "Thought": { "type": "string", "minLength": 1, "maxLength": 2000 },
            "Answer": { "type": "string", "minLength": 1, "maxLength": 200 }
        }
    }
    """
    Thought: str = pydantic.Field(
        ...,
        min_length=1,
        max_length=2000,
        description="The model's thinking process"
    )
    Answer: str = pydantic.Field(
        ...,
        min_length=1,
        max_length=200,
        description="The model's final answer"
    )

MODEL_TEMPLATES: Dict[str, Type[pydantic.BaseModel]] = {
    "ner": NerModel,
    "triples": TriplesModel,
    "fact": FactModel,
    "json": ArbitraryJsonModel,
    "qa_cot": QaCotModel,
}

def get_pydantic_model(template_name: str) -> Type[pydantic.BaseModel]:
    """
    A factory function that returns the corresponding Pydantic model class based on the template name (key).

    Args:
        template_name: The name of the template (e.g., 'ner', 'triples').

    Returns:
        The corresponding Pydantic.BaseModel subclass.

    Raises:
        ValueError: If the provided template_name is invalid.
    """
    model_class = MODEL_TEMPLATES.get(template_name)
    if model_class is None:
        available_keys = ", ".join(MODEL_TEMPLATES.keys())
        raise ValueError(f"Unknown template name: '{template_name}'. Available templates: {available_keys}")
    
    return model_class



import os

from ..utils.logging_utils import get_logger
from ..utils.config_utils import BaseConfig

from .openai_gpt import CacheOpenAI
from .base import BaseLLM
from .bedrock_llm import BedrockLLM
from .transformers_llm import TransformersLLM

__all__ = ["CacheOpenAI", "BaseLLM", "BedrockLLM", "TransformersLLM"]

logger = get_logger(__name__)


def _get_llm_class(config: BaseConfig):
    if config.llm_base_url is not None and 'localhost' in config.llm_base_url and os.getenv('OPENAI_API_KEY') is None:
        os.environ['OPENAI_API_KEY'] = 'sk-'

    if config.llm_name.startswith('bedrock'):
        return BedrockLLM(config)
    
    if config.llm_name.startswith('Transformers/'):
        return TransformersLLM(config)
    
    return CacheOpenAI.from_experiment_config(config)



"""
异步OpenIE模块

提供异步的信息抽取功能，包括：
1. 异步NER（命名实体识别）
2. 异步三元组抽取
3. 异步批量处理
4. 智能重试和错误处理

性能优化：
- 使用asyncio实现真正的非阻塞I/O
- 并发控制和速率限制
- 智能缓存减少API调用
- 批量处理优化
"""

import asyncio
import json
import re
import logging
from typing import Dict, Any, List, Tuple, Optional
from dataclasses import dataclass
from tqdm.asyncio import tqdm_asyncio

from ..utils.async_utils import (
    AsyncOpenAIClient, 
    AsyncBatchProcessor,
    AsyncRateLimiter,
    AsyncCircuitBreaker,
    async_retry,
    async_timeout,
    global_async_monitor
)
from ..utils.misc_utils import TripleRawOutput, NerRawOutput
from ..utils.llm_utils import fix_broken_generated_json, filter_invalid_triples
from ..prompts import PromptTemplateManager

logger = logging.getLogger(__name__)


@dataclass
class AsyncOpenIEConfig:
    """异步OpenIE配置"""
    max_concurrent_ner: int = 20
    max_concurrent_triples: int = 20
    ner_timeout: float = 60.0
    triple_timeout: float = 60.0
    max_retries: int = 3
    cache_enabled: bool = True
    rate_limit_rps: float = 50.0


class AsyncOpenIE:
    """异步OpenIE处理器"""
    
    def __init__(
        self,
        api_key: str,
        base_url: Optional[str] = None,
        model: str = "gpt-4o-mini",
        config: AsyncOpenIEConfig = None,
        cache_dir: str = "outputs/async_cache"
    ):
        self.api_key = api_key
        self.base_url = base_url
        self.model = model
        self.config = config or AsyncOpenIEConfig()
        
        self.client = AsyncOpenAIClient(
            api_key=api_key,
            base_url=base_url,
            cache_dir=cache_dir
        )
        
        self.prompt_template_manager = PromptTemplateManager(
            role_mapping={"system": "system", "user": "user", "assistant": "assistant"}
        )
        
        self.ner_processor = AsyncBatchProcessor(
            max_concurrent=self.config.max_concurrent_ner,
            timeout=self.config.ner_timeout
        )
        
        self.triple_processor = AsyncBatchProcessor(
            max_concurrent=self.config.max_concurrent_triples,
            timeout=self.config.triple_timeout
        )
        
        self.rate_limiter = AsyncRateLimiter(
            requests_per_second=self.config.rate_limit_rps,
            burst_size=int(self.config.rate_limit_rps * 2)
        )
        
        self.circuit_breaker = AsyncCircuitBreaker(
            failure_threshold=5,
            recovery_timeout=30.0
        )
    
    def _extract_ner_from_response(self, response: str) -> List[str]:
        """从响应中提取命名实体"""
        pattern = r'\{[^{}]*"named_entities"\s*:\s*\[[^\]]*\][^{}]*\}'
        match = re.search(pattern, response, re.DOTALL)
        if match is None:
            return []
        try:
            return eval(match.group())["named_entities"]
        except Exception:
            return []
    
    def _extract_triples_from_response(self, response: str) -> List[Tuple]:
        """从响应中提取三元组"""
        pattern = r'\{[^{}]*"triples"\s*:\s*\[[^\]]*\][^{}]*\}'
        match = re.search(pattern, response, re.DOTALL)
        if match is None:
            return []
        try:
            return eval(match.group())["triples"]
        except Exception:
            return []
    
    @async_retry(max_retries=3, delay=1.0, backoff=2.0)
    @async_timeout(60.0)
    async def ner(self, chunk_key: str, passage: str) -> NerRawOutput:
        """异步NER"""
        async with global_async_monitor.record_time("async_ner"):
            await self.rate_limiter.acquire()
            
            ner_input_message = self.prompt_template_manager.render(
                name='ner', 
                passage=passage
            )
            
            try:
                raw_response, metadata, cache_hit = await self.circuit_breaker.call(
                    self.client.chat_completion,
                    messages=ner_input_message,
                    model=self.model
                )
                
                metadata['cache_hit'] = cache_hit
                
                if metadata.get('finish_reason') == 'length':
                    real_response = fix_broken_generated_json(raw_response)
                else:
                    real_response = raw_response
                
                extracted_entities = self._extract_ner_from_response(real_response)
                unique_entities = list(dict.fromkeys(extracted_entities))
                
                return NerRawOutput(
                    chunk_id=chunk_key,
                    response=raw_response,
                    unique_entities=unique_entities,
                    metadata=metadata
                )
                
            except Exception as e:
                logger.warning(f"NER failed for chunk {chunk_key}: {e}")
                return NerRawOutput(
                    chunk_id=chunk_key,
                    response="",
                    unique_entities=[],
                    metadata={'error': str(e)}
                )
    
    @async_retry(max_retries=3, delay=1.0, backoff=2.0)
    @async_timeout(60.0)
    async def triple_extraction(
        self, 
        chunk_key: str, 
        passage: str, 
        named_entities: List[str]
    ) -> TripleRawOutput:
        """异步三元组抽取"""
        async with global_async_monitor.record_time("async_triple_extraction"):
            await self.rate_limiter.acquire()
            
            messages = self.prompt_template_manager.render(
                name='triple_extraction',
                passage=passage,
                named_entity_json=json.dumps({"named_entities": named_entities})
            )
            
            try:
                raw_response, metadata, cache_hit = await self.circuit_breaker.call(
                    self.client.chat_completion,
                    messages=messages,
                    model=self.model
                )
                
                metadata['cache_hit'] = cache_hit
                
                if metadata.get('finish_reason') == 'length':
                    real_response = fix_broken_generated_json(raw_response)
                else:
                    real_response = raw_response
                
                extracted_triples = self._extract_triples_from_response(real_response)
                triplets = filter_invalid_triples(triples=extracted_triples)
                
                return TripleRawOutput(
                    chunk_id=chunk_key,
                    response=raw_response,
                    metadata=metadata,
                    triples=triplets
                )
                
            except Exception as e:
                logger.warning(f"Triple extraction failed for chunk {chunk_key}: {e}")
                return TripleRawOutput(
                    chunk_id=chunk_key,
                    response="",
                    metadata={'error': str(e)},
                    triples=[]
                )
    
    async def openie(self, chunk_key: str, passage: str) -> Dict[str, Any]:
        """异步OpenIE（NER + 三元组抽取）"""
        ner_output = await self.ner(chunk_key=chunk_key, passage=passage)
        triple_output = await self.triple_extraction(
            chunk_key=chunk_key,
            passage=passage,
            named_entities=ner_output.unique_entities
        )
        return {"ner": ner_output, "triplets": triple_output}
    
    async def batch_ner(
        self, 
        chunk_passages: Dict[str, str],
        show_progress: bool = True
    ) -> Dict[str, NerRawOutput]:
        """异步批量NER"""
        async def process_item(item: Tuple[str, str]) -> NerRawOutput:
            chunk_key, passage = item
            return await self.ner(chunk_key, passage)
        
        items = list(chunk_passages.items())
        
        if show_progress:
            results = []
            for coro in tqdm_asyncio.as_completed(
                [process_item(item) for item in items],
                total=len(items),
                desc="Async NER"
            ):
                result = await coro
                results.append(result)
        else:
            results = await self.ner_processor.process_batch(
                items,
                process_item,
                desc="Async NER"
            )
        
        return {res.chunk_id: res for res in results if res is not None}
    
    async def batch_triple_extraction(
        self,
        chunk_passages: Dict[str, str],
        ner_results: Dict[str, NerRawOutput],
        show_progress: bool = True
    ) -> Dict[str, TripleRawOutput]:
        """异步批量三元组抽取"""
        async def process_item(item: Tuple[str, NerRawOutput]) -> TripleRawOutput:
            chunk_key, ner_result = item
            return await self.triple_extraction(
                chunk_key=chunk_key,
                passage=chunk_passages[chunk_key],
                named_entities=ner_result.unique_entities
            )
        
        items = list(ner_results.items())
        
        if show_progress:
            results = []
            for coro in tqdm_asyncio.as_completed(
                [process_item(item) for item in items],
                total=len(items),
                desc="Async Triple Extraction"
            ):
                result = await coro
                results.append(result)
        else:
            results = await self.triple_processor.process_batch(
                items,
                process_item,
                desc="Async Triple Extraction"
            )
        
        return {res.chunk_id: res for res in results if res is not None}
    
    async def batch_openie(
        self,
        chunks: Dict[str, Dict[str, Any]],
        show_progress: bool = True
    ) -> Tuple[Dict[str, NerRawOutput], Dict[str, TripleRawOutput]]:
        """
        异步批量OpenIE
        
        Args:
            chunks: chunk_id -> {content: str, ...} 的映射
            show_progress: 是否显示进度条
            
        Returns:
            (ner_results_dict, triple_results_dict)
        """
        chunk_passages = {k: v["content"] for k, v in chunks.items()}
        
        logger.info(f"Starting async batch OpenIE for {len(chunks)} chunks")
        
        ner_results = await self.batch_ner(chunk_passages, show_progress)
        
        triple_results = await self.batch_triple_extraction(
            chunk_passages, ner_results, show_progress
        )
        
        logger.info(f"Async batch OpenIE completed: {len(ner_results)} NER, {len(triple_results)} triples")
        
        return ner_results, triple_results
    
    async def close(self):
        """关闭客户端"""
        await self.client.close()
    
    async def get_stats(self) -> Dict[str, Any]:
        """获取统计信息"""
        return await global_async_monitor.get_all_stats()


class AsyncOpenIEWithFallback:
    """带回退机制的异步OpenIE"""
    
    def __init__(
        self,
        async_openie: AsyncOpenIE,
        sync_openie: Any
    ):
        self.async_openie = async_openie
        self.sync_openie = sync_openie
    
    async def ner(self, chunk_key: str, passage: str) -> NerRawOutput:
        """异步NER（带回退）"""
        try:
            return await self.async_openie.ner(chunk_key, passage)
        except Exception as e:
            logger.warning(f"Async NER failed, falling back to sync: {e}")
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(
                None,
                self.sync_openie.ner,
                chunk_key,
                passage
            )
    
    async def triple_extraction(
        self,
        chunk_key: str,
        passage: str,
        named_entities: List[str]
    ) -> TripleRawOutput:
        """异步三元组抽取（带回退）"""
        try:
            return await self.async_openie.triple_extraction(
                chunk_key, passage, named_entities
            )
        except Exception as e:
            logger.warning(f"Async triple extraction failed, falling back to sync: {e}")
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(
                None,
                self.sync_openie.triple_extraction,
                chunk_key,
                passage,
                named_entities
            )
    
    async def batch_openie(
        self,
        chunks: Dict[str, Dict[str, Any]],
        show_progress: bool = True
    ) -> Tuple[Dict[str, NerRawOutput], Dict[str, TripleRawOutput]]:
        """异步批量OpenIE（带回退）"""
        try:
            return await self.async_openie.batch_openie(chunks, show_progress)
        except Exception as e:
            logger.warning(f"Async batch OpenIE failed, falling back to sync: {e}")
            loop = asyncio.get_event_loop()
            return await loop.run_in_executor(
                None,
                self.sync_openie.batch_openie,
                chunks
            )
    
    async def close(self):
        """关闭客户端"""
        await self.async_openie.close()



"""
异步I/O优化模块

提供全面的异步I/O优化，包括：
1. 异步LLM调用（带连接池和重试）
2. 异步嵌入编码
3. 异步缓存操作
4. 异步批量处理
5. 异步检索操作

性能优化策略：
- 使用asyncio实现真正的非阻塞I/O
- 连接池复用减少连接开销
- 信号量控制并发数避免过载
- 批量操作减少网络往返
- 智能重试机制提高可靠性
"""

import asyncio
import hashlib
import json
import os
import sqlite3
import time
import logging
from typing import Any, Dict, List, Optional, Tuple, Callable, TypeVar, ParamSpec, Awaitable
from dataclasses import dataclass
from functools import wraps
from collections import OrderedDict

# 延迟导入重型库以提高启动速度
_aiohttp = None
_numpy = None

def _get_aiohttp():
    """延迟导入 aiohttp"""
    global _aiohttp
    if _aiohttp is None:
        import aiohttp
        _aiohttp = aiohttp
    return _aiohttp

def _get_numpy():
    """延迟导入 numpy"""
    global _numpy
    if _numpy is None:
        import numpy
        _numpy = numpy
    return _numpy



logger = logging.getLogger(__name__)

P = ParamSpec('P')
T = TypeVar('T')


@dataclass
class AsyncConfig:
    """异步配置"""
    max_concurrent_requests: int = 100
    max_connections_per_host: int = 50
    connection_timeout: float = 30.0
    read_timeout: float = 300.0
    max_retries: int = 3
    retry_delay: float = 1.0
    retry_backoff: float = 2.0
    cache_enabled: bool = True
    cache_max_memory_size: int = 1000


class AsyncLRUCache:
    """异步线程安全的LRU缓存"""
    
    def __init__(self, max_size: int = 1000):
        self.max_size = max_size
        self._cache: OrderedDict = OrderedDict()
        self._lock = asyncio.Lock()
        self._stats = {'hits': 0, 'misses': 0}
    
    async def get(self, key: str) -> Optional[Any]:
        async with self._lock:
            if key in self._cache:
                self._cache.move_to_end(key)
                self._stats['hits'] += 1
                return self._cache[key]
            self._stats['misses'] += 1
            return None
    
    async def put(self, key: str, value: Any) -> None:
        async with self._lock:
            if key in self._cache:
                self._cache.move_to_end(key)
            else:
                if len(self._cache) >= self.max_size:
                    self._cache.popitem(last=False)
            self._cache[key] = value
    
    async def delete(self, key: str) -> bool:
        async with self._lock:
            if key in self._cache:
                del self._cache[key]
                return True
            return False
    
    async def clear(self) -> None:
        async with self._lock:
            self._cache.clear()
            self._stats = {'hits': 0, 'misses': 0}
    
    async def get_stats(self) -> Dict[str, Any]:
        async with self._lock:
            total = self._stats['hits'] + self._stats['misses']
            return {
                'size': len(self._cache),
                'max_size': self.max_size,
                'hits': self._stats['hits'],
                'misses': self._stats['misses'],
                'hit_rate': self._stats['hits'] / total if total > 0 else 0
            }


class AsyncSQLiteCache:
    """异步SQLite缓存（使用线程池执行）- 懒加载优化"""
    
    def __init__(self, cache_dir: str, cache_name: str = "async_cache.sqlite"):
        os.makedirs(cache_dir, exist_ok=True)
        self.cache_file = os.path.join(cache_dir, cache_name)
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._initialized = False  # 懒加载标记
        self._init_lock = asyncio.Lock()
    
    async def _ensure_initialized(self):
        """确保数据库已初始化（懒加载）"""
        if self._initialized:
            return
        
        async with self._init_lock:
            if self._initialized:
                return
            
            loop = self._get_loop()
            
            def _sync_init():
                conn = sqlite3.connect(self.cache_file)
                c = conn.cursor()
                c.execute("""
                    CREATE TABLE IF NOT EXISTS async_cache (
                        key TEXT PRIMARY KEY,
                        value TEXT,
                        namespace TEXT,
                        created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                        expires_at TIMESTAMP
                    )
                """)
                c.execute("CREATE INDEX IF NOT EXISTS idx_namespace ON async_cache(namespace)")
                c.execute("CREATE INDEX IF NOT EXISTS idx_expires ON async_cache(expires_at)")
                conn.commit()
                conn.close()
            
            await loop.run_in_executor(None, _sync_init)
            self._initialized = True
    
    def _get_loop(self) -> asyncio.AbstractEventLoop:
        """获取事件循环"""
        if self._loop is None:
            self._loop = asyncio.get_event_loop()
        return self._loop
    
    async def get(self, key: str, namespace: str = "default") -> Optional[Any]:
        """异步获取缓存"""
        await self._ensure_initialized()  # 懒加载初始化
        loop = self._get_loop()
        
        def _sync_get():
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute(
                "SELECT value FROM async_cache WHERE key = ? AND namespace = ?",
                (key, namespace)
            )
            row = c.fetchone()
            conn.close()
            return row[0] if row else None
        
        result = await loop.run_in_executor(None, _sync_get)
        return json.loads(result) if result else None
    
    async def put(self, key: str, value: Any, namespace: str = "default", 
                  ttl_seconds: Optional[int] = None) -> None:
        """异步存储缓存"""
        await self._ensure_initialized()  # 懒加载初始化
        loop = self._get_loop()
        
        def _sync_put():
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            expires_at = None
            if ttl_seconds:
                expires_at = f"datetime('now', '+{ttl_seconds} seconds')"
            c.execute(
                """INSERT OR REPLACE INTO async_cache (key, value, namespace, expires_at) 
                   VALUES (?, ?, ?, """ + (expires_at if expires_at else "NULL") + """)""",
                (key, json.dumps(value, ensure_ascii=False), namespace)
            )
            conn.commit()
            conn.close()
        
        await loop.run_in_executor(None, _sync_put)
    
    async def get_batch(self, keys: List[str], namespace: str = "default") -> Dict[str, Any]:
        """异步批量获取缓存"""
        loop = self._get_loop()
        
        def _sync_batch_get():
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            placeholders = ','.join(['?' for _ in keys])
            c.execute(
                f"SELECT key, value FROM async_cache WHERE key IN ({placeholders}) AND namespace = ?",
                keys + [namespace]
            )
            rows = c.fetchall()
            conn.close()
            return {row[0]: json.loads(row[1]) for row in rows}
        
        return await loop.run_in_executor(None, _sync_batch_get)
    
    async def put_batch(self, items: Dict[str, Any], namespace: str = "default",
                        ttl_seconds: Optional[int] = None) -> None:
        """异步批量存储缓存"""
        loop = self._get_loop()
        
        def _sync_batch_put():
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            for key, value in items.items():
                c.execute(
                    "INSERT OR REPLACE INTO async_cache (key, value, namespace) VALUES (?, ?, ?)",
                    (key, json.dumps(value, ensure_ascii=False), namespace)
                )
            conn.commit()
            conn.close()
        
        await loop.run_in_executor(None, _sync_batch_put)
    
    async def delete(self, key: str, namespace: str = "default") -> bool:
        """异步删除缓存"""
        loop = self._get_loop()
        
        def _sync_delete():
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute(
                "DELETE FROM async_cache WHERE key = ? AND namespace = ?",
                (key, namespace)
            )
            deleted = c.rowcount > 0
            conn.commit()
            conn.close()
            return deleted
        
        return await loop.run_in_executor(None, _sync_delete)
    
    async def clear_expired(self) -> int:
        """清理过期缓存"""
        loop = self._get_loop()
        
        def _sync_clear():
            conn = sqlite3.connect(self.cache_file)
            c = conn.cursor()
            c.execute("DELETE FROM async_cache WHERE expires_at IS NOT NULL AND expires_at < datetime('now')")
            deleted = c.rowcount
            conn.commit()
            conn.close()
            return deleted
        
        return await loop.run_in_executor(None, _sync_clear)


class AsyncHTTPClient:
    """异步HTTP客户端（带连接池和重试）"""
    
    def __init__(self, config: AsyncConfig = None):
        if config is None:
            config = AsyncConfig()
        self.config = config
        self._semaphore = asyncio.Semaphore(config.max_concurrent_requests)
        self._session: Optional[Any] = None  # aiohttp.ClientSession
    
    async def get_session(self) -> Any:
        """获取或创建会话（延迟加载 aiohttp）"""
        aiohttp = _get_aiohttp()
        if self._session is None or self._session.closed:
            timeout = aiohttp.ClientTimeout(
                total=self.config.read_timeout,
                connect=self.config.connection_timeout
            )
            connector = aiohttp.TCPConnector(
                limit=self.config.max_connections_per_host,
                limit_per_host=self.config.max_connections_per_host,
                enable_cleanup_closed=True
            )
            self._session = aiohttp.ClientSession(
                timeout=timeout,
                connector=connector
            )
        return self._session
    
    async def close(self):
        """关闭会话"""
        if self._session is not None and not self._session.closed:
            await self._session.close()
            self._session = None
    
    async def request(
        self, 
        method: str, 
        url: str, 
        **kwargs
    ) -> Tuple[Any, Dict[str, Any]]:
        """发送请求（带重试和限流）"""
        aiohttp = _get_aiohttp()
        async with self._semaphore:
            session = await self.get_session()
            
            last_error = None
            for attempt in range(self.config.max_retries):
                try:
                    async with session.request(method, url, **kwargs) as response:
                        if response.status == 200:
                            data = await response.json()
                            return data, {'status': response.status}
                        elif response.status in [429, 503]:
                            retry_after = float(response.headers.get('Retry-After', self.config.retry_delay))
                            await asyncio.sleep(retry_after * (self.config.retry_backoff ** attempt))
                            continue
                        else:
                            text = await response.text()
                            raise aiohttp.ClientResponseError(
                                request_info=response.request_info,
                                history=response.history,
                                status=response.status,
                                message=text
                            )
                except aiohttp.ClientError as e:
                    last_error = e
                    if attempt < self.config.max_retries - 1:
                        await asyncio.sleep(self.config.retry_delay * (self.config.retry_backoff ** attempt))
                    continue
                except asyncio.TimeoutError as e:
                    last_error = e
                    if attempt < self.config.max_retries - 1:
                        await asyncio.sleep(self.config.retry_delay * (self.config.retry_backoff ** attempt))
                    continue
            
            raise last_error or Exception(f"Request failed after {self.config.max_retries} retries")
    
    async def get(self, url: str, **kwargs) -> Tuple[Any, Dict[str, Any]]:
        return await self.request('GET', url, **kwargs)
    
    async def post(self, url: str, **kwargs) -> Tuple[Any, Dict[str, Any]]:
        return await self.request('POST', url, **kwargs)
    
    async def request_batch(
        self, 
        requests: List[Dict[str, Any]]
    ) -> List[Tuple[Any, Dict[str, Any]]]:
        """批量发送请求"""
        tasks = [
            self.request(req['method'], req['url'], **req.get('kwargs', {}))
            for req in requests
        ]
        return await asyncio.gather(*tasks, return_exceptions=True)


class AsyncOpenAIClient:
    """异步OpenAI客户端"""
    
    def __init__(
        self,
        api_key: str,
        base_url: Optional[str] = None,
        config: AsyncConfig = None,
        cache_dir: str = "outputs/async_cache",
        # 支持单独的嵌入 API 配置
        embedding_api_key: Optional[str] = None,
        embedding_base_url: Optional[str] = None,
    ):
        self.api_key = api_key
        self.base_url = base_url or "https://api.openai.com/v1"
        self.config = config or AsyncConfig()
        self.http_client = AsyncHTTPClient(self.config)
        self.memory_cache = AsyncLRUCache(max_size=self.config.cache_max_memory_size)
        self.persistent_cache = AsyncSQLiteCache(cache_dir, "openai_async_cache.sqlite")
        
        self.headers = {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json"
        }
        
        # 嵌入 API 配置
        self.embedding_api_key = embedding_api_key or api_key
        self.embedding_base_url = embedding_base_url or self.base_url
        self.embedding_headers = {
            "Authorization": f"Bearer {self.embedding_api_key}",
            "Content-Type": "application/json"
        }
    
    def _compute_cache_key(self, messages: List[Dict], model: str, **kwargs) -> str:
        """计算缓存键"""
        content = json.dumps({
            'messages': messages,
            'model': model,
            **{k: v for k, v in kwargs.items() if k in ['temperature', 'max_tokens', 'seed']}
        }, sort_keys=True)
        return hashlib.md5(content.encode()).hexdigest()
    
    async def chat_completion(
        self,
        messages: List[Dict],
        model: str = "gpt-4o-mini",
        **kwargs
    ) -> Tuple[str, Dict[str, Any], bool]:
        """异步聊天补全"""
        cache_key = self._compute_cache_key(messages, model, **kwargs)
        
        if self.config.cache_enabled:
            cached = await self.memory_cache.get(cache_key)
            if cached is not None:
                return cached['content'], cached['metadata'], True
            
            cached = await self.persistent_cache.get(cache_key, "chat_completion")
            if cached is not None:
                await self.memory_cache.put(cache_key, cached)
                return cached['content'], cached['metadata'], True
        
        url = f"{self.base_url}/chat/completions"
        payload = {
            "model": model,
            "messages": messages,
            **kwargs
        }
        
        data, resp_meta = await self.http_client.post(
            url,
            headers=self.headers,
            json=payload
        )
        
        content = data['choices'][0]['message']['content']
        metadata = {
            'prompt_tokens': data['usage']['prompt_tokens'],
            'completion_tokens': data['usage']['completion_tokens'],
            'model': data['model'],
            'finish_reason': data['choices'][0]['finish_reason']
        }
        
        if self.config.cache_enabled:
            cache_value = {'content': content, 'metadata': metadata}
            await self.memory_cache.put(cache_key, cache_value)
            await self.persistent_cache.put(cache_key, cache_value, "chat_completion")
        
        return content, metadata, False
    
    async def chat_completion_batch(
        self,
        batch_messages: List[List[Dict]],
        model: str = "gpt-4o-mini",
        **kwargs
    ) -> List[Tuple[str, Dict[str, Any], bool]]:
        """异步批量聊天补全"""
        tasks = [
            self.chat_completion(messages, model, **kwargs)
            for messages in batch_messages
        ]
        return await asyncio.gather(*tasks, return_exceptions=True)
    
    async def embedding(
        self,
        texts: List[str],
        model: str = "text-embedding-3-small"
    ) -> Tuple[Any, Dict[str, Any]]:
        """异步嵌入编码"""
        np = _get_numpy()
        cache_key = hashlib.md5(json.dumps({'texts': texts, 'model': model}).encode()).hexdigest()
        
        if self.config.cache_enabled:
            cached = await self.memory_cache.get(cache_key)
            if cached is not None:
                return np.array(cached['embeddings']), cached['metadata']
        
        # 使用单独的嵌入 API 配置
        url = f"{self.embedding_base_url}/embeddings"
        payload = {
            "model": model,
            "input": [t.replace("\n", " ") for t in texts]
        }
        
        data, resp_meta = await self.http_client.post(
            url,
            headers=self.embedding_headers,
            json=payload
        )
        
        embeddings = np.array([item['embedding'] for item in data['data']])
        metadata = {
            'model': data['model'],
            'total_tokens': data['usage']['total_tokens']
        }
        
        if self.config.cache_enabled:
            await self.memory_cache.put(cache_key, {
                'embeddings': embeddings.tolist(),
                'metadata': metadata
            })
        
        return embeddings, metadata
    
    async def embedding_batch(
        self,
        all_texts: List[List[str]],
        model: str = "text-embedding-3-small",
        batch_size: int = 100
    ) -> List[Tuple[Any, Dict[str, Any]]]:
        """异步批量嵌入编码"""
        tasks = []
        for texts in all_texts:
            for i in range(0, len(texts), batch_size):
                batch = texts[i:i + batch_size]
                tasks.append(self.embedding(batch, model))
        
        results = await asyncio.gather(*tasks, return_exceptions=True)
        return results
    
    async def close(self):
        """关闭客户端"""
        await self.http_client.close()


class AsyncBatchProcessor:
    """异步批量处理器"""
    
    def __init__(
        self,
        max_concurrent: int = 50,
        batch_size: int = 10,
        timeout: float = 300.0
    ):
        self.max_concurrent = max_concurrent
        self.batch_size = batch_size
        self.timeout = timeout
        self._semaphore = asyncio.Semaphore(max_concurrent)
    
    async def process_batch(
        self,
        items: List[Any],
        process_fn: Callable[[Any], Awaitable[T]],
        desc: str = "Processing"
    ) -> List[T]:
        """异步批量处理"""
        results = [None] * len(items)
        
        async def process_with_semaphore(idx: int, item: Any):
            async with self._semaphore:
                try:
                    result = await asyncio.wait_for(
                        process_fn(item),
                        timeout=self.timeout
                    )
                    return idx, result, None
                except Exception as e:
                    return idx, None, e
        
        tasks = [
            process_with_semaphore(i, item)
            for i, item in enumerate(items)
        ]
        
        completed = 0
        for coro in asyncio.as_completed(tasks):
            idx, result, error = await coro
            completed += 1
            if error:
                logger.error(f"Batch processing error at index {idx}: {error}")
                results[idx] = None
            else:
                results[idx] = result
            
            if completed % self.batch_size == 0:
                logger.info(f"{desc}: {completed}/{len(items)} completed")
        
        return results
    
    async def process_in_chunks(
        self,
        items: List[Any],
        process_fn: Callable[[List[Any]], Awaitable[List[T]]],
        chunk_size: int = None
    ) -> List[T]:
        """分块批量处理"""
        if chunk_size is None:
            chunk_size = self.batch_size
        
        chunks = [items[i:i + chunk_size] for i in range(0, len(items), chunk_size)]
        
        results = []
        for i, chunk in enumerate(chunks):
            try:
                chunk_results = await asyncio.wait_for(
                    process_fn(chunk),
                    timeout=self.timeout
                )
                results.extend(chunk_results)
                logger.info(f"Processed chunk {i + 1}/{len(chunks)}")
            except Exception as e:
                logger.error(f"Error processing chunk {i}: {e}")
                results.extend([None] * len(chunk))
        
        return results


class AsyncRateLimiter:
    """异步速率限制器"""
    
    def __init__(
        self,
        requests_per_second: float = 10.0,
        burst_size: int = 20
    ):
        self.requests_per_second = requests_per_second
        self.burst_size = burst_size
        self._tokens = burst_size
        self._last_update = time.monotonic()
        self._lock = asyncio.Lock()
    
    async def acquire(self, tokens: int = 1):
        """获取令牌"""
        async with self._lock:
            now = time.monotonic()
            elapsed = now - self._last_update
            self._tokens = min(
                self.burst_size,
                self._tokens + elapsed * self.requests_per_second
            )
            self._last_update = now
            
            if self._tokens < tokens:
                wait_time = (tokens - self._tokens) / self.requests_per_second
                await asyncio.sleep(wait_time)
                self._tokens = 0
            else:
                self._tokens -= tokens


class AsyncCircuitBreaker:
    """异步熔断器"""
    
    def __init__(
        self,
        failure_threshold: int = 5,
        recovery_timeout: float = 60.0,
        expected_exception: type = Exception
    ):
        self.failure_threshold = failure_threshold
        self.recovery_timeout = recovery_timeout
        self.expected_exception = expected_exception
        
        self._failure_count = 0
        self._last_failure_time: Optional[float] = None
        self._state = 'closed'  # closed, open, half_open
        self._lock = asyncio.Lock()
    
    async def call(self, func: Callable[..., Awaitable[T]], *args, **kwargs) -> T:
        """调用函数（带熔断保护）"""
        async with self._lock:
            if self._state == 'open':
                if time.monotonic() - self._last_failure_time > self.recovery_timeout:
                    self._state = 'half_open'
                else:
                    raise Exception("Circuit breaker is open")
        
        try:
            result = await func(*args, **kwargs)
            async with self._lock:
                if self._state == 'half_open':
                    self._state = 'closed'
                self._failure_count = 0
            return result
        except self.expected_exception:
            async with self._lock:
                self._failure_count += 1
                self._last_failure_time = time.monotonic()
                if self._failure_count >= self.failure_threshold:
                    self._state = 'open'
            raise


def async_retry(
    max_retries: int = 3,
    delay: float = 1.0,
    backoff: float = 2.0,
    exceptions: tuple = (Exception,)
):
    """异步重试装饰器"""
    def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            last_exception = None
            for attempt in range(max_retries):
                try:
                    return await func(*args, **kwargs)
                except exceptions as e:
                    last_exception = e
                    if attempt < max_retries - 1:
                        wait_time = delay * (backoff ** attempt)
                        logger.warning(f"Retry {attempt + 1}/{max_retries} after {wait_time}s: {e}")
                        await asyncio.sleep(wait_time)
            raise last_exception
        return wrapper
    return decorator


def async_timeout(seconds: float):
    """异步超时装饰器"""
    def decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return await asyncio.wait_for(func(*args, **kwargs), timeout=seconds)
        return wrapper
    return decorator


async def run_in_threadpool(func: Callable[..., T], *args, **kwargs) -> T:
    """在线程池中运行同步函数"""
    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, lambda: func(*args, **kwargs))


async def gather_with_concurrency(n: int, *tasks, return_exceptions: bool = False):
    """限制并发数的gather"""
    semaphore = asyncio.Semaphore(n)
    
    async def sem_task(task):
        async with semaphore:
            return await task
    
    return await asyncio.gather(
        *[sem_task(task) for task in tasks],
        return_exceptions=return_exceptions
    )


class AsyncPerformanceMonitor:
    """异步性能监控器"""
    
    def __init__(self):
        self.metrics: Dict[str, List[Dict[str, Any]]] = {}
        self._lock = asyncio.Lock()
    
    async def record(self, name: str, value: float, metadata: Dict[str, Any] = None):
        """记录指标"""
        async with self._lock:
            if name not in self.metrics:
                self.metrics[name] = []
            self.metrics[name].append({
                'value': value,
                'timestamp': time.time(),
                'metadata': metadata or {}
            })
    
    def record_time(self, name: str):
        """记录时间的上下文管理器"""
        class AsyncTimer:
            def __init__(self, monitor, metric_name):
                self.monitor = monitor
                self.metric_name = metric_name
                self.start_time = None
            
            async def __aenter__(self):
                self.start_time = time.time()
                return self
            
            async def __aexit__(self, *args):
                elapsed = time.time() - self.start_time
                await self.monitor.record(self.metric_name, elapsed)
        
        return AsyncTimer(self, name)
    
    async def get_stats(self, name: str) -> Dict[str, float]:
        """获取统计信息"""
        async with self._lock:
            if name not in self.metrics:
                return {}
            
            values = [m['value'] for m in self.metrics[name]]
            if not values:
                return {}
            
            # 使用内置函数计算统计信息，避免导入 numpy
            count = len(values)
            mean = sum(values) / count
            variance = sum((x - mean) ** 2 for x in values) / count
            std = variance ** 0.5
            
            return {
                'count': count,
                'mean': mean,
                'std': std,
                'min': min(values),
                'max': max(values),
                'total': sum(values)
            }
    
    async def get_all_stats(self) -> Dict[str, Dict[str, float]]:
        """获取所有统计信息"""
        async with self._lock:
            result = {}
            for name in self.metrics:
                result[name] = await self.get_stats(name)
            return result
    
    async def clear(self):
        """清除所有指标"""
        async with self._lock:
            self.metrics.clear()


global_async_monitor = AsyncPerformanceMonitor()



from .ner import one_shot_ner_paragraph, one_shot_ner_output
from ...utils.llm_utils import convert_format_to_template

ner_conditioned_re_system = """Your task is to construct an RDF (Resource Description Framework) graph from the given passages and named entity lists. 
Respond with a JSON list of triples, with each triple representing a relationship in the RDF graph. 

Pay attention to the following requirements:
- Each triple should contain at least one, but preferably two, of the named entities in the list for each passage.
- Clearly resolve pronouns to their specific names to maintain clarity.

"""


ner_conditioned_re_frame = """Convert the paragraph into a JSON dict, it has a named entity list and a triple list.
Paragraph:
```
{passage}
```

{named_entity_json}
"""


ner_conditioned_re_input = ner_conditioned_re_frame.format(passage=one_shot_ner_paragraph, named_entity_json=one_shot_ner_output)


ner_conditioned_re_output = """{"triples": [
            ["Radio City", "located in", "India"],
            ["Radio City", "is", "private FM radio station"],
            ["Radio City", "started on", "3 July 2001"],
            ["Radio City", "plays songs in", "Hindi"],
            ["Radio City", "plays songs in", "English"],
            ["Radio City", "forayed into", "New Media"],
            ["Radio City", "launched", "PlanetRadiocity.com"],
            ["PlanetRadiocity.com", "launched in", "May 2008"],
            ["PlanetRadiocity.com", "is", "music portal"],
            ["PlanetRadiocity.com", "offers", "news"],
            ["PlanetRadiocity.com", "offers", "videos"],
            ["PlanetRadiocity.com", "offers", "songs"]
    ]
}
"""


prompt_template = [
    {"role": "system", "content": ner_conditioned_re_system},
    {"role": "user", "content": ner_conditioned_re_input},
    {"role": "assistant", "content": ner_conditioned_re_output},
    {"role": "user", "content": convert_format_to_template(original_string=ner_conditioned_re_frame, placeholder_mapping=None, static_values=None)}
]


ner_system = """Your task is to extract named entities from the given paragraph. 
Respond with a JSON list of entities.
"""

one_shot_ner_paragraph = """Radio City
Radio City is India's first private FM radio station and was started on 3 July 2001.
It plays Hindi, English and regional songs.
Radio City recently forayed into New Media in May 2008 with the launch of a music portal - PlanetRadiocity.com that offers music related news, videos, songs, and other music-related features."""


one_shot_ner_output = """{"named_entities":
    ["Radio City", "India", "3 July 2001", "Hindi", "English", "May 2008", "PlanetRadiocity.com"]
}
"""


prompt_template = [
    {"role": "system", "content": ner_system},
    {"role": "user", "content": one_shot_ner_paragraph},
    {"role": "assistant", "content": one_shot_ner_output},
    {"role": "user", "content": "${passage}"}
]


from .Contriever import ContrieverModel
from .base import EmbeddingConfig, BaseEmbeddingModel
from .GritLM import GritLMEmbeddingModel
from .NVEmbedV2 import NVEmbedV2EmbeddingModel
from .OpenAI import OpenAIEmbeddingModel
from .Cohere import CohereEmbeddingModel
from .Transformers import TransformersEmbeddingModel
from .VLLM import VLLMEmbeddingModel

from ..utils.logging_utils import get_logger

__all__ = [
    "ContrieverModel",
    "EmbeddingConfig",
    "BaseEmbeddingModel",
    "GritLMEmbeddingModel",
    "NVEmbedV2EmbeddingModel",
    "OpenAIEmbeddingModel",
    "CohereEmbeddingModel",
    "TransformersEmbeddingModel",
    "VLLMEmbeddingModel",
]

logger = get_logger(__name__)


def _get_embedding_model_class(embedding_model_name: str = "nvidia/NV-Embed-v2"):
    if "GritLM" in embedding_model_name:
        return GritLMEmbeddingModel
    elif "NV-Embed-v2" in embedding_model_name:
        return NVEmbedV2EmbeddingModel
    elif "contriever" in embedding_model_name:
        return ContrieverModel
    elif "text-embedding" in embedding_model_name:
        return OpenAIEmbeddingModel
    elif "cohere" in embedding_model_name:
        return CohereEmbeddingModel
    elif embedding_model_name.startswith("Transformers/"):
        return TransformersEmbeddingModel
    elif embedding_model_name.startswith("VLLM/"):
        return VLLMEmbeddingModel
    # 支持更多OpenAI兼容的嵌入模型
    elif "bge" in embedding_model_name.lower():
        return OpenAIEmbeddingModel
    elif "embedding" in embedding_model_name.lower():
        return OpenAIEmbeddingModel
    # 默认使用OpenAI兼容模式（大多数API都兼容）
    else:
        logger.warning(f"Unknown embedding model name: {embedding_model_name}, using OpenAI compatible mode")
        return OpenAIEmbeddingModel
